# Example 2.7 from
#
# Near-Optimal Robust Bilevel Optimization
#   M. Besancon, M. F. Anjos and L. Brotcorne
#   arXiv:1908.04040v5 (2019)
#
# Optimal solution: (x,v) = (0,1)
#
import pyomo.environ as pe
from pao.pyomo import *


def create():
    M = pe.ConcreteModel()

    M.x = pe.Var(bounds=(0,None))
    M.v = pe.Var()

    M.o = pe.Objective(expr=M.x, sense=pe.minimize)
    M.c = pe.Constraint(expr= M.v >= 1 - M.x/10)

    M.L = SubModel(fixed=M.x)
    M.L.o = pe.Objective(expr=M.v, sense=pe.maximize)
    M.L.c = pe.Constraint(expr= 1 + M.x/10 >= M.v)

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    #M.pprint()
    opt = Solver("pao.pyomo.FA")
    opt.solve(M)
    print(pe.value(M.x))
    print(pe.value(M.v))



#
# Classes used to define a solver API
#
import os
import copy
import six
import abc
import enum
import textwrap
import logging

from pyutilib.misc import Options

import pyomo.opt.parallel.manager
import pyomo.environ as pe
from pyomo.common.config import ConfigValue, ConfigBlock, add_docstring_list
from pyomo.neos.kestrel import kestrelAMPL

__all__ = ['TerminationCondition', 'SolverAPI', 'Results', 'Solver']


_logger = logging.getLogger('pyomo')


class TerminationCondition(enum.Enum):
    """
    The TerminationCondition class defines a enumeration of optimization termination conditions.
    """

    unknown = 0
    """unknown serves as both a default value, and it is used when no other enum member makes sense"""

    maxTimeLimit = 1
    """The solver exited due to a time limit"""

    maxIterations = 2
    """The solver exited due to an iteration limit """

    objectiveLimit = 3
    """The solver exited due to an objective limit"""

    minStepLength = 4
    """The solver exited due to a minimum step length"""

    optimal = 5
    """The solver exited with an optimal solution"""

    locallyOptimal = 6
    """The solver exited with a locally optimal solution"""

    globallyOptimal = 7
    """The solver exited with a locally optimal solution"""

    unbounded = 8
    """The solver exited because the problem is unbounded"""

    infeasible = 9
    """The solver exited because the problem is infeasible"""

    infeasibleOrUnbounded = 10
    """The solver exited because the problem is either infeasible or unbounded"""

    error = 11
    """The solver exited due to an error"""

    interrupted = 12
    """The solver exited because it was interrupted"""

    licensingProblems = 13
    """The solver exited due to licensing problems"""


class SolverAPI(abc.ABC):
    """
    The base class for all PAO solvers.

    The SolverAPI class defines a consistent API for optimization solvers.
    """

    config = ConfigBlock()
    #config.declare('time_limit', ConfigValue(
    #    default=None,
    #    domain=float,
    #    description="Solver time limit (in seconds)",
    #    ))
    #config.declare('keep_files', ConfigValue(
    #    default=False,
    #    domain=bool,
    #    description="If True, then temporary files are not deleted. (default is False)",
    #    ))
    config.declare('tee', ConfigValue(
        default=False,
        domain=bool,
        description="If True, then solver output is streamed to stdout. (default is False)",
        ))
    config.declare('load_solutions', ConfigValue(
        default=True,
        domain=bool,
        description="If True, then the finale solution is loaded into the model. (default is True)",
        ))

    def __init__(self):
        # Create a per-instance copy of the configuration data
        self.config = self.config()

    def available(self):
        """
        Returns a bool indicating if the solver can be executed.

        The default behavior is to always return `True`, but this method
        can be overloaded in a subclass to support solver-specific logic
        (e.g.  to confirm that a solver license is available).

        Returns
        -------
        bool
            This method returns True if the solver can be executed.
        """
        True

    def valid_license(self):
        """
        Returns a bool indicating if the solver has a valid license.

        The default behavior is to always return `True`, but this method
        can be overloaded in a subclass to support solver-specific logic
        (e.g.  to check the solver license).

        Returns
        -------
        bool
            This method returns True if the solver license is valid.
        """
        True

    def version(self):
        """
        Returns a tuple that describes the solver version.

        The return value is a tuple of strings.  A typical format is (major, minor, patch), but this
        is not required. The default behavior is to return an empty tuple.

        Returns
        -------
        tuple
            The solver version.
        """
        return tuple()

    @abc.abstractmethod
    def solve(self, model, **options):
        """
        Executes the solver and loads the solution into the model.

        Parameters
        ----------
        model
            The model that is being optimized.
        options
            Keyword options that are used to configure the solver.

        {}
        Returns
        -------
        Results
            A summary of the optimization results.
        """
        pass

    __solve_doc__ = solve.__doc__

    @staticmethod
    def _generate_solve_docstring(cls):
        """
        Generate the docstring for cls.solve, including the description
        of the keyword arguments defined by a pyomo configuration object.

        Parameters
        ----------
        cls:
            The subclass of SolverAPI whose docstring is being generated
        """
        cls.solve.__doc__ = SolverAPI.__solve_doc__.format( add_docstring_list("", cls.config, 8) )

    def is_persistent():
        """
        Returns True if the solver is persistent.

        Persistent solvers maintain the model representation in memory,
        which enables performance optimization when a problem is resolved
        after changing initial conditions or tweaking model parameters.

        The default is to return False, but this method can be overloaded
        in a subclass to support solver-specific logic.

        Returns
        -------
        bool
            This method returns True if the solver is persistent.
        """
        return False

    def __bool__(self):
        """
        Raises an error because this class cannot be interpreted with a boolean.

        Raises
        ------
        RuntimeError
            Casting a solver to bool is not allowed.
        """
        raise RuntimeError("Casting a solver to bool() is not allowed.  Use available() to check if the solver can be executed.")

    def __enter__(self):
        """
        Setup a solver and return it within a context manager.

        Returns
        -------
        SolverAPI
            Return a reference to **self**.
        """
        return self

    def __exit__(self, t, v, traceback):
        """
        Cleanup the solver at the end of a context manager.
        """
        pass

    def _update_config(self, config_options, config=None, validate_options=True):
        """
        .. todo::
            Document where and how this is used.
        """
        if config is None:
            config = self.config
        keys = set(config_options.keys())
        for k,v in config_options.items():
            if k in config:
                config[k] = v
                keys.remove(k)    
        if validate_options:
            assert (len(keys) == 0), "Unexpected options to solve() have been specified: %s" % " ".join(sorted(k for k in keys))
        return {key:config_options[key] for key in keys}

SolverAPI._generate_solve_docstring(SolverAPI)


class PyomoSolver(SolverAPI):

    config = SolverAPI.config()
    config.declare('executable', ConfigValue(
        default=None,
        description="The path to the executable used for this solver."
        ))

    def __init__(self, name, options):
        # Create a per-instance copy of the configuration data
        self.config = self.config()
        self.name = name
        level = _logger.getEffectiveLevel()
        _logger.setLevel(logging.ERROR)
        self.solver = pe.SolverFactory(name)
        _logger.setLevel(level)
        self.solver_options = self._update_config(options, validate_options=False)
        self.solver.options.update(self.solver_options)
        if self.config['executable'] is not None:
            self.solver.set_executable(self.config['executable'])

    def available(self):
        if self.config['executable'] is not None:
            self.solver.set_executable(self.config['executable'])
        return self.solver.available(exception_flag=False)

    def solve(self, model, **options):
        assert (isinstance(model, pe.Model) or isinstance(model, pe.SimpleBlock)), "The Pyomo solver '%s' cannot solve a model of type %s" % (self.name, str(type(model)))
        tmp_config = self.config()
        tmp_options = self._update_config(options, config=tmp_config, validate_options=False)
        if tmp_config['executable'] is not None:
            self.solver.set_executable(tmp_config['executable'])
        return self.solver.solve(model, tee=tmp_config.tee, options=tmp_options)


class NEOSSolver(SolverAPI):

    config = SolverAPI.config()
    config.declare('server', ConfigValue(
        default=None,
        description="The server that is used to execute the solver."
        ))
    config.declare('email', ConfigValue(
        default=None,
        description="The email that NEOS requires for use.  If not specified, the NEOS_EMAIL environment variable is used."
        ))

    def __init__(self, name, options):
        # Create a per-instance copy of the configuration data
        self.config = self.config()
        self.name = name
        self.solver_options = self._update_config(options, validate_options=False)
        self.neos_available = None

    def available(self):
        if self.neos_available is None:
            self.neos_available = False
            try:
                if kestrelAMPL().neos is not None:
                    self.neos_available = True
            except:
                pass
        return self.neos_available

    def solve(self, model, **options):
        assert (isinstance(model, pe.Model) or isinstance(model, pe.SimpleBlock)), "The Pyomo solver '%s' cannot solve a model of type %s" % (self.name, str(type(model)))
        if self.config['email'] is not None:
            os.environ['NEOS_EMAIL'] = self.config['email']
        assert ('NEOS_EMAIL' in os.environ), "The NEOS solver requires an email.  Please specify the NEOS_EMAIL environment variable."
        solver_manager = pe.SolverManagerFactory('neos')
        try:
            tmp_config = self.config()
            tmp_options = copy.copy(self.solver_options)
            tmp_options.update( self._update_config(options, config=tmp_config, validate_options=False) )
            results = solver_manager.solve(model, opt=self.name, tee=tmp_config.tee, options=tmp_options)
            return results
        except pyomo.opt.parallel.manager.ActionManagerError as err:
            raise RuntimeError(str(err)) from None
        

"""
>>> opt = Solver('my_solver')
>>> results = opt.solve(my_model, load_solutions=False)
>>> if results.solver.termination_condition == TerminationCondition.optimal:
>>>     print('optimal solution found: ', results.solver.best_feasible_objective)
>>>     results.load_solution( my_model )
>>>     print('the optimal value of x is ', my_model.x.value)
>>> elif results.found_feasible_solution():
>>>     print('sub-optimal but feasible solution found: ', results.solver.best_feasible_objective)
>>>     results.load_vars(vars_to_load=[my_model.x])
>>>     print('The value of x in the feasible solution is ', my_model.x.value)
>>> elif results.solver.termination_condition in {TerminationCondition.maxIterations,
...                                               TerminationCondition.maxTimeLimit}:
>>>     print('No feasible solution was found. The best lower bound found was ',
...           results.solver.best_objective_bound)
>>> else:
>>>     print('The following termination condition was encountered: ',
...           results.solver.termination_condition)
"""
class ResultsBase(abc.ABC):
    """
    Defines the API for results objects.

    A Results object is returned from this method.  This object
    reports information from a solver's execution, including the
    termination condition, objective value for the best feasible point
    and the best bound computed on the objective value.  Additionally,
    a results object may contain solutions found during optimization.

    Attributes
    ----------
    solution_manager
        An object that manages storing and loading data to and from models.
    """

    def __init__(self):
        self.solution_manager = None
        self.solver = Options(
                termination_condition=TerminationCondition.unknown,
                best_feasible_objective=None,
                )
        self.problem = Options()

    @abc.abstractmethod
    def found_feasible_solution(self):
        """
        Returns
        -------
        bool
            True if at least one feasible solution was found. False otherwise.
        """
        pass

    def store_to(self, model, i=0):
        """
        Store the solution in this object into the given model.

        A results object may contain one or more solutions. This method
        copies the **i**-th solution into the model.

        Parameters
        ----------
        i: int
            The index of the solution copied into the model.
        """
        if self.solution_manager is None:
            raise RuntimeError("The results object is missing a solution manager.")
        self.solution_manager.store_to(model, i=0)

    def load_from(self, model):
        """
        Load solution from the model.

        When completed, this results object contains only one solution, which corresponds
        to the solution in the model.
        """
        if self.solution_manager is None:
            raise RuntimeError("The results object is missing a solution manager.")
        self.solution_manager.load_from(model)

    @abc.abstractmethod
    def __str__(self):
        """
        Generate a string summary of this results object.

        Returns
        -------
        str
            A string summarizing the data in this object.
        """
        pass


class Results(ResultsBase):
    """
    The results object.
    """

    def __init__(self, found_feasible_solution=None):
        super(Results, self).__init__()
        self._found_feasible_solution = found_feasible_solution

    def found_feasible_solution(self):
        return self._found_feasible_solution

    def check_optimal_termination(self):
        """
        This function returns True if the termination condition for the solver
        is 'optimal', 'locallyOptimal', or 'globallyOptimal', and the status is 'ok'

        Returns
        -------
        bool
        """
        if self.solver.termination_condition == TerminationCondition.optimal \
            or self.solver.termination_condition == TerminationCondition.locallyOptimal \
            or self.solver.termination_condition == TerminationCondition.globallyOptimal:
            return True
        return False

    def __str__(self):
        problem = self.problem.__str__(indent='  ')
        solver = self.solver.__str__(indent='  ')
        return "Problem:\n-" + problem[1:] + \
            "\nSolver:\n-" + solver[1:]


class SolverFactory(object):
    """
    A class that manages a registry of solvers.

    A solver factory manages a registry that enables 
    solvers to be created by name.
    """

    _registry = {}
    _doc = {}

    def register(self, cls=None, *, name=None, doc=None):
        """
        Register a solver with the specified name.

        Parameters
        ----------
        cls
            Class type for the solver
        name: str
            Unique name of the solver
        doc: str
            Short description of the solver

        Returns
        -------
        decorator
            If the **cls** parameter is None, then a class 
            decorator function
            is returned that can be used to register a solver.
        """
        def decorator(cls):
            assert (name is not None), "Must register a solver with a name"
            #assert (name in SolverFactory._registry), "Name '%s' is already registered!" % name
            SolverFactory._registry[name] = cls
            SolverFactory._doc[name] = doc
            return cls
        if cls is None:
            return decorator
        return decorator(cls)

    def __iter__(self):
        """
        Yields
        -------
        string
            Yields the next solver name that has been registered
        """
        for name in sorted(SolverFactory._doc.keys()):
            yield name

    def summary(self):
        """
        Print a summary of all solvers.
        """
        for name in self:
            print(name)
            print(textwrap.indent("\n".join(textwrap.wrap(self.description(name))), "    "))
            print("")

    def description(self, name):
        """
        Returns the description of the specified solver.

        Parameters
        ----------
        name: str
            The name of a solver

        Returns
        -------
        str
            A short description of the specified solver
        """
        assert (name in SolverFactory._registry), "Unknown solver '%s' specified" % name
        return SolverFactory._doc[name]

    def __call__(self, name, **options):
        """
        Constructs the specified solver.

        This method creates a class instance for the solver that is specified.

        Parameters
        ----------
        name: str
            The name of a solver
        options
            Keyword options that are used to configure the solver.

        Returns
        -------
        SolverAPI
            A solver class instance for the solver that is specified.
        """
        #
        # Create a solver registered in PAO
        #
        if name in SolverFactory._registry:
            solver = SolverFactory._registry[name]()
            solver._update_config(options)
            return solver
        #
        # Create a NEOS solver interface
        #
        server = options.pop("server",None)
        if server == 'neos':
            solver = NEOSSolver(name, options)
            assert (solver.available()), "NEOS is not available.  Cannot use NEOS solver '%s'." % name
            return solver
        assert (server is None), "Unknown solver server: "+server
        #
        # Create an solver interface using Pyomo, but
        # fail if the solver is not available.
        #
        solver = PyomoSolver(name, options)
        assert (solver.available()), "Unknown solver '%s' specified" % name
        return solver


Solver = SolverFactory()
"""
Solver is a global instance of the :class:`SolverFactory`.
This object provides a global registry of PAO optimization solvers.
"""



#
# A stochastic bilevel interdiction adapted from:
#   "A Stochastic Program for Interdicting Smuggled Nuclear Material"
#   by F. Pan, W. S. Charlton, D. P. Morton
#   DRAFT August 19, 2002
#
# example contributor: She'ifa Punla-Green
#
import pyomo.environ as pe
from pao.pyomo import *

#A incidence matrix
#n number of nodes
#e number of edges
#b budget
#c cost of edge sensor placement
#ns number of starting nodes
#nt number of ending nodes

def create():
    b = 1
    e = 16
    n = 12
    ns = 3
    nt = 4

    E = [(2, 1), (3, 1), (3, 2), (3, 6), (4, 3), (4, 5), (5, 2), (5, 6), (6, 4), (6, 7), (6, 8), (6, 9), (9, 11), (10, 6),
         (11, 10), (11, 12)]
    W = [(4, 1), (4, 7), (4, 8), (4, 12), (5, 1), (5, 7), (5, 8), (5, 12), (11, 1), (11, 7), (11, 8), (11, 12)]
    FS = {}
    FS[1] = []
    FS[2] = [(2, 1)]
    FS[3] = [(3, 1), (3, 2), (3, 6)]
    FS[4] = [(4, 3), (4, 5)]
    FS[5] = [(5, 2), (5, 6)]
    FS[6] = [(6, 4), (6, 7), (6, 8), (6, 9)]
    FS[7] = []
    FS[8] = []
    FS[9] = [(9, 11)]
    FS[10] = [(10, 6)]
    FS[11] = [(11, 10), (11, 12)]
    FS[12] = []

    RS = {}
    RS[1] = [(2, 1), (3, 1)]
    RS[2] = [(3, 2), (5, 2)]
    RS[3] = [(4, 3)]
    RS[4] = [(6, 4)]
    RS[5] = [(4, 5)]
    RS[6] = [(3, 6), (5, 6), (10, 6)]
    RS[7] = [(6, 7)]
    RS[8] = [(6, 8)]
    RS[9] = [(6, 9)]
    RS[10] = [(11, 10)]
    RS[11] = [(9, 11)]
    RS[12] = [(11, 12)]

    P = {}
    P[(4, 1)] = 1 / (ns * nt)
    P[(4, 7)] = 1 / (ns * nt)
    P[(4, 8)] = 1 / (ns * nt)
    P[(4, 12)] = 1 / (ns * nt)
    P[(5, 1)] = 1 / (ns * nt)
    P[(5, 7)] = 1 / (ns * nt)
    P[(5, 8)] = 1 / (ns * nt)
    P[(5, 12)] = 1 / (ns * nt)
    P[(11, 1)] = 1 / (ns * nt)
    P[(11, 7)] = 1 / (ns * nt)
    P[(11, 8)] = 1 / (ns * nt)
    P[(11, 12)] = 1 / (ns * nt)

    c = 1

    p = {}
    p[(2, 1)] = 0.5
    p[(3, 1)] = 0.5
    p[(3, 2)] = 0.5
    p[(3, 6)] = 0.5
    p[(4, 3)] = 0.25
    p[(4, 5)] = 0.25
    p[(5, 2)] = 0.25
    p[(5, 6)] = 0.25
    p[(6, 4)] = 0.25
    p[(6, 7)] = 0.5
    p[(6, 8)] = 0.5
    p[(6, 9)] = 0.5
    p[(9, 11)] = 0.25
    p[(10, 6)] = 0.5
    p[(11, 10)] = 0.25
    p[(11, 12)] = 0.25

    q = {}
    q[(2, 1)] = 0.25
    q[(3, 1)] = 0.25
    q[(3, 2)] = 0.25
    q[(3, 6)] = 0.25
    q[(4, 3)] = 0.1
    q[(4, 5)] = 0.1
    q[(5, 2)] = 0.1
    q[(5, 6)] = 0.1
    q[(6, 4)] = 0.1
    q[(6, 7)] = 0.25
    q[(6, 8)] = 0.25
    q[(6, 9)] = 0.25
    q[(9, 11)] = 0.1
    q[(10, 6)] = 0.25
    q[(11, 10)] = 0.1
    q[(11, 12)] = 0.1

    m = pe.ConcreteModel()
    m.nodeset = pe.RangeSet(1, n)
    m.edgeset = pe.RangeSet(1, e)
    m.Edgeset = pe.Set(initialize=E)
    m.Omega = pe.RangeSet(1, nt*ns)
    m.OmegaSet = pe.Set(initialize=W)
    m.Node = pe.Block(m.nodeset)

    for j in m.nodeset:
        m.Node[j].FS = pe.Set(initialize=FS[j])
        m.Node[j].RS = pe.Set(initialize=RS[j])

    m.b=pe.Param(initialize=b)
    m.c=pe.Param(m.Edgeset,initialize=c)
    m.P=pe.Param(m.OmegaSet,initialize=P)
    m.p=pe.Param(m.Edgeset,initialize=p)
    m.q=pe.Param(m.Edgeset,initialize=q)
    m.x=pe.Var(m.Edgeset,within=pe.Binary)
    m.h=pe.Var(m.OmegaSet,within=pe.NonNegativeReals)

    def Obj(m):
        value=sum(m.P[i,j]*m.h[i,j] for (i,j) in m.OmegaSet) #h[i,j] is the objective of subproblem (i,j)
        return value
    m.Obj = pe.Objective(rule=Obj, sense=pe.minimize)

    def C(m):
        value=sum(m.c[i,j]*m.x[i,j] for (i,j) in m.Edgeset)
        return value <= b
    m.C=pe.Constraint(rule=C)


    m.Sub = pe.Block(m.OmegaSet)
    for (s,t) in m.OmegaSet:
        m.Sub[s,t].sub = SubModel(fixed=m.x)
        m.Sub[s,t].sub.stset = pe.Set(initialize=[s,t])
        m.Sub[s,t].sub.y = pe.Var(m.Edgeset,within=pe.NonNegativeReals, bounds=(0,1))
        m.Sub[s,t].sub.z = pe.Var(m.Edgeset,within=pe.NonNegativeReals, bounds=(0,1))
        def obj(sub):
            return m.h[s,t]
        m.Sub[(s,t)].sub.obj = pe.Objective(rule=obj, sense=pe.maximize)

        def cb(sub):
            value=sum(m.Sub[(s,t)].sub.y[(i,j)]+m.Sub[(s,t)].sub.z[(i,j)] for (i,j) in m.Node[s].FS)
            return value==1
        m.Sub[(s,t)].sub.cb=pe.Constraint(rule=cb)

        def cc(sub,i):
            value=sum(m.Sub[(s,t)].sub.y[i,j]+m.Sub[(s,t)].sub.z[(i,j)] for (i,j) in m.Node[i].FS)
            value=value - sum(m.p[(j,i)]*m.Sub[(s,t)].sub.y[(j,i)]+m.q[(j,i)]*m.Sub[(s,t)].sub.z[(j,i)] for (j,i) in m.Node[i].RS)
            return value == 0
        m.Sub[(s,t)].sub.cc=pe.Constraint(m.nodeset-m.Sub[(s,t)].sub.stset,rule=cc)

        def cd(sub):
            value = sum (m.p[(j,i)]*m.Sub[(s,t)].sub.y[(j,i)]+m.q[(j,i)]*m.Sub[(s,t)].sub.z[(j,i)] for (j,i) in m.Node[t].RS)
            return m.h[(s,t)]-value == 0
        m.Sub[(s,t)].sub.cd=pe.Constraint(rule=cd)

        def ce(sub,i,j):
            return m.Sub[(s,t)].sub.y[(i,j)] <= 1- m.x[(i,j)]
        m.Sub[(s,t)].sub.ce=pe.Constraint(m.Edgeset,rule=ce)

        def cf(sub,i,j):
            return m.Sub[(s,t)].sub.z[(i,j)] <= m.x[(i,j)]
        m.Sub[(s,t)].sub.cf=pe.Constraint(m.Edgeset,rule=cf)

    m.weights = {m.Sub[k].name+'.sub':v for k,v in m.P.items()}

    return m

if __name__ == "__main__":
    M = create()

    opt = Solver('pao.pyomo.FA')
    opt.solve(M, solver='gurobi', tee=True)

    print(M.x)
    M.x.pprint()
    print(M.h)
    M.h.pprint()

    for s,t in M.OmegaSet:
        print(M.Sub[s,t])
        M.Sub[s,t].sub.y.pprint()
        M.Sub[s,t].sub.z.pprint()

    #    self.assertTrue(math.isclose(M.xR.value, 2))
    #    self.assertTrue(math.isclose(M.L.xR.value, 100))



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

"""
pao.pyomo.plugins.highpoint
"""

import six

from pyomo.core.kernel.component_map import ComponentMap
from pyomo.core import Block, Objective,\
                       Var, Constraint, ComponentUID, \
                       TransformationFactory, Reference, SimpleVar
from pao.pyomo.components import SubModel
from .transform import BaseBilevelTransformation
import logging
from pyomo.core.kernel.component_map import ComponentMap

logger = logging.getLogger(__name__)


def create_submodel_hp_block(instance):
    """
    Creates highpoint relaxation with the given specified model; does
    not include any submodel or block that is deactivated
    """
    block = Block(concrete=True)

    # create a component map to keep track of id() for the
    # original object and the referenced object
    block._map = ComponentMap()

    # get the objective for the master problem
    for c in instance.component_objects(Objective, descend_into=False):
        ref = Reference(c)
        block.add_component(c.name, ref)
        block._map[c] = ref[None]

    # get the variables of the model (if there are more submodels, then
    # extraneous variables may be added to the block)
    for c in instance.component_objects(Var, sort=True, descend_into=True, active=True):
        if c.is_indexed():
            ref = Reference(c[...])
            block.add_component(c.name, ref)
            block._map[c] = ref
        else:
            ref = Reference(c)
            block.add_component(c.name, ref)
            block._map[c] = ref[None]

    # get the constraints from the main model
    for c in instance.component_objects(Constraint, sort=True, descend_into=True, active=True):
        if c.is_indexed():
            ref = Reference(c[...])
            block.add_component(c.name, ref)
            block._map[c] = ref
        else:
            ref = Reference(c)
            block.add_component(c.name, ref)
            block._map[c] = ref[None]

    # deactivate the highpoint relaxation
    # block.deactivate()

    return block


@TransformationFactory.register('pao.pyomo.highpoint',
                                doc="Generate a highpoint relaxation of the model")
class LinearHighpointTransformation(BaseBilevelTransformation):
    """
    This transformation creates a block using a SubModel object,
    which contains objective and constraint set of upper-level, and constraint set of lower-level for
    lower-level feasibility.
    """

    def _apply_to(self, model, **kwds):
        submodel_name = kwds.pop('submodel_name', 'hpr')
        self._preprocess('pao.pyomo.highpoint', model)

        map = ComponentMap()
        for key, sub in self.submodel.items():
            model.reclassify_component_type(sub, Block)

        setattr(model, submodel_name,
                    create_submodel_hp_block(model))

        model._transformation_data['pao.pyomo.highpoint'].submodel_cuid = \
            ComponentUID(model)
        model._transformation_data['pao.pyomo.highpoint'].block_cuid = \
            ComponentUID(getattr(model, submodel_name))

        for key, sub in self.submodel.items():
            model.reclassify_component_type(sub, SubModel)



#
# Example from Pineda and Morales (2018), showing how
# bad estimates of bigMs give the wrong answer.
#
from pao.mpr import *


def create():
    M = LinearMultilevelProblem()

    U = M.add_upper(nxR=1)
    L = U.add_lower(nxR=1)

    U.maximize = True
    U.x.lower_bounds = [0]
    U.x.upper_bounds = [2]
    U.c[U] = [1]
    U.c[L] = [1]

    L.x.lower_bounds = [0]
    L.c[L] = [1]

    L.A[U] = [[100]]
    L.A[L] = [[-1]]
    L.b = [100]

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    opt = Solver('pao.mpr.FA')
    opt.solve(M)
    M.print()



#
# Simple interdiction problem solvable through
# lower level dualization and MPEC transformation
#
# Note that this is a quadratic bilevel problem
#
# This model has an optimal objective value of 0
# To see this, fix u and optimize the lower level
#
# DUALIZED:
# min   v1 + v2 + u
# s.t.  u1 + v1 >= 1    : x
#       -u*u1 + v2 >= 0  : y
#       v1, v2, >= 0
#       u1 unconstrained
#

from pao.mpr import *

def create():
    M = QuadraticMultilevelProblem()

    U = M.add_upper(nxB=1)
    L = U.add_lower(nxR=2)
    L.maximize = True

    L.x.lower_bounds=[0, 0]
    L.x.upper_bounds=[1, 1]

    U.c[U] = [1]
    U.c[L] = [1, 0]

    L.c[U] = [1]
    L.c[L] = [1, 0]

    L.A[L] = [[ 1, 0],
              [-1, 0]]
    L.Q[U,L] = (2,1,2), {(0,0,1):-1, (1,0,1):1}
    L.b    = [0, 0]

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    m = linearize_bilinear_terms(M)
    m.print()

    opt = Solver('pao.mpr.FA')
    opt.solve(m, tee=True)
    m.print()



#
# A solver interface to MibS
#
# A branch-and-cut algorithm for mixed integer bilevel linear optimization problems and its implementation
# S Tahernejad, TK Ralphs, ST DeNegre
# Mathematical Programming Computation 12 (4), 529-568
#
# TODO - Citations
#
import os
import sys
import time
import numpy as np
import pyutilib
import pyomo.environ as pe
import pyomo.opt
from pyomo.common.config import ConfigBlock, ConfigValue
#from pyomo.mpec import ComplementarityList, complements

import pao.common
from ..solver import Solver, LinearMultilevelSolverBase, LinearMultilevelResults
from ..repn import LinearMultilevelProblem
from ..convert_repn import convert_to_standard_form
from . import pyomo_util
#from .reg import create_model_replacing_LL_with_kkt


@Solver.register(
        name='pao.mpr.MIBS',
        doc='PAO solver for Multilevel Problem Representations using the COIN-OR MibS solver by Tahernejad, Ralphs, and DeNegre (2020).')

class LinearMultilevelSolver_MIBS(LinearMultilevelSolverBase):
    """
    PAO MibS solver for linear MPRs: pao.mpr.MIBS
    """
    config = LinearMultilevelSolverBase.config()
    config.declare('executable', ConfigValue(
        default='mibs',
        description="The executable used for MibS.  (default is mibs)"
        ))
    config.declare('param_file', ConfigValue(
        default=None,
        description="The parameter file used to configure MibS.  (default is None)"
        ))

    def __init__(self, **kwds):
        super().__init__(name='pao.mpr.MIBS')

    def check_model(self, model):
        #
        # Confirm that the LinearMultilevelProblem is well-formed
        #
        assert (type(model) is LinearMultilevelProblem), "Solver '%s' can only solve a linear multilevel problem" % self.name
        model.check()
        #
        # Confirm that this is a bilevel problem with just one lower-level
        #
        for L in model.U.LL:
            assert (len(L.LL) == 0), "Can only solve bilevel problems"
        assert (len(model.U.LL) == 1), "Can only solve a bilevel problem with a single lower-level"

    def solve(self, model, **options):
        #
        # Error checks
        #
        self.check_model(model)
        #
        # Process keyword options
        #
        self._update_config(options)
        #
        # Start clock
        #
        start_time = time.time()

        self.standard_form, soln_manager = convert_to_standard_form(model, inequalities=True)

        #
        # Write the MPS file and MIBS auxilliary file
        #
        # TODO - Make these temporary files
        #
        M = self.create_mibs_model(model, "mibs.mps", "mibs.aux")

        cmd = [ self.config['executable'], '-Alps_instance', 'mibs.mps', '-MibS_auxiliaryInfoFile', 'mibs.aux']
        if self.config['param_file'] is not None:
            cmd.append('-param')
            cmd.append('mibs.par')

        ans = pao.common.run_shellcmd(cmd, tee=self.config['tee'])
        os.remove("mibs.mps")
        os.remove("mibs.aux")
        #print("RC", ans.rc)
        #print("LOG", ans.log)

        results = self._initialize_results(ans, model)
        results.check_optimal_termination()

        results.solver.wallclock_time = time.time() - start_time
        return results

    def _initialize_results(self, ans, M):
        #
        # Default value is zero
        #
        M.U.x.values = [0]*len(M.U.x)
        M.U.LL.x.values = [0]*len(M.U.LL.x)
        #
        # Results
        #
        results = pao.common.Results()
        solv = results.solver
        solv.termination_condition = pao.common.TerminationCondition.unknown
        solv.name = self.config['executable']
        solv.rc = ans.rc
        #
        # Parse Log
        #
        # TODO - Handle errors
        #
        if ans.rc == 0:
            state = 0
            for line in ans.log.split("\n"):
                #print(state, line)
                line = line.strip()
                if state == 0:
                    if line.startswith("Optimal solution:"):
                        solv.termination_condition = pao.common.TerminationCondition.optimal
                        state = 1
                elif state == 1:
                    solv.best_feasible_objective = float(line.split("=")[1])
                    state = 2
                elif state == 2:
                    if line.startswith("Number"):
                        state = 3
                        continue
                    name, _, val = line.split(" ")
                    vname, index = name[:-1].split("[")
                    #print(vname, index, val)
                    if vname == "x":
                        M.U.x.values[int(index)] = float(val)
                    else:
                        M.U.LL.x.values[int(index)] = float(val)
        return results

    def _debug(self, M):    # pragma: no cover
        for j in M.U.xR:
            print("U",j,pe.value(M.U.xR[j]))
        for j in M.L.xR:
            print("L",j,pe.value(M.L.xR[j]))
        for j in M.kkt.lam:
            print("lam",j,pe.value(M.kkt.lam[j]))
        for j in M.kkt.nu:
            print("nu",j,pe.value(M.kkt.nu[j]))





    def create_mibs_model(self, repn, mps_filename, aux_filename):
        """
        TODO - Document this transformation
        """
        U = repn.U
        L = repn.U.LL[0]

        #---------------------------------------------------
        # Create Pyomo model
        #---------------------------------------------------

        M = pe.ConcreteModel()
        M.U = pe.Block()
        M.L = pe.Block()

        # upper-level variables
        pyomo_util.add_variables(M.U, U)
        # lower-level variables
        pyomo_util.add_variables(M.L, L)

        # objective
        e = pyomo_util.dot(U.c[U], U.x, num=1) + U.d
        e += pyomo_util.dot(U.c[L], L.x, num=1)
        M.o = pe.Objective(expr=e)

        # upper-level constraints
        pyomo_util.add_linear_constraints(M.U, U.A, U, L, U.b, U.inequalities)
        # lower-level constraints
        pyomo_util.add_linear_constraints(M.L, L.A, U, L, L.b, L.inequalities)

        #---------------------------------------------------
        # Write files
        #---------------------------------------------------
        # TODO - get variable mapping information
        #
        M.write("tmp.mps")
        #
        # Add a space after "BOUND"
        #
        with open("tmp.mps", "r") as INPUT:
            with open(mps_filename, "w") as OUTPUT:
                for line in INPUT:
                    OUTPUT.write(line.replace("BOUND ", "BOUND  "))
        os.remove("tmp.mps")

        with open(aux_filename, "w") as OUTPUT:
            # Num lower-level variables
            OUTPUT.write("N {}\n".format(len(L.x)))
            # Num lower-level constraints
            OUTPUT.write("M {}\n".format(L.b.size))
            # Indices of lower-level variables
            nx_upper = len(U.x)
            for i in range(len(L.x)):
                OUTPUT.write("LC {}\n".format(i+nx_upper))
            # Indices of lower-level constraints
            nc_upper = U.b.size
            for i in range(L.b.size):
                OUTPUT.write("LR {}\n".format(i+nc_upper))
            # Coefficients for lower-level objective
            for i in range(len(L.x)):
                OUTPUT.write("LO {}\n".format(L.c[L][i]))
            # Lower-level objective sense
            if L.minimize:
                OUTPUT.write("OS 1\n")
            else:
                OUTPUT.write("OS -1\n")
        
        return M

pao.common.SolverAPI._generate_solve_docstring(LinearMultilevelSolver_MIBS)



from .repn import LinearMultilevelProblem, QuadraticMultilevelProblem
from . import convert_repn
from . import examples
from .convert_repn import linearize_bilinear_terms
from .solver import Solver
#from . import pyomo_solvers
from . import solvers
from . import soln_manager



#
# Example from
# Moore, J. and J. Bard 1990.
# The mixed integer linear bilevel programming problem.
# Operations Research 38(5), 911–921.
#
from pao.mpr import *


def create():
    M = LinearMultilevelProblem()

    U = M.add_upper(nxZ=1)
    L = U.add_lower(nxZ=1)

    U.c[U] = [-1]
    U.c[L] = [-10]

    L.c[L] = [1]

    L.A[U] = [[-25],
              [1],
              [2],
              [-2]]
    L.A[L] = [[20],
              [2],
              [-1],
              [-10]]
    L.b = [30,10,15,-15]

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    M.print()
    opt = Solver('pao.mpr.FA')
    opt.solve(M)



#
# Toy Example 3 from 
#
# "A projection-based reformulation and decomposition algorithm for global optimization 
#  of a class of mixed integer bilevel linear programs" by Dajun Yue, Jiyao Gao, Bo Zeng, Fengqi You
#
import pyomo.environ as pe
from pao.pyomo import *


def create():
    M = pe.ConcreteModel()

    M.xR = pe.Var(bounds=(0,None))
    M.xZ = pe.Var(bounds=(0,None), within=pe.Integers)

    M.L = SubModel(fixed=(M.xR, M.xZ))
    M.L.xR = pe.Var(bounds=(0,None))
    M.L.xZ = pe.Var(bounds=(0,None), within=pe.Integers)

    M.o = pe.Objective(expr=20*M.xR - 38*M.xZ + M.L.xR + 42*M.L.xZ, sense=pe.minimize)
    M.c1 = pe.Constraint(expr=7*M.xR + 5*M.xZ +             7*M.L.xZ <= 62)
    M.c2 = pe.Constraint(expr=6*M.xR + 9*M.xZ + 10*M.L.xR + 2*M.L.xZ <= 117)

    M.L.o = pe.Objective(expr=39*M.L.xR + 27*M.L.xZ, sense=pe.maximize)
    M.L.c1 = pe.Constraint(expr=8*M.xR + 2*M.L.xR + 8*M.L.xZ <= 53)
    M.L.c2 = pe.Constraint(expr=9*M.xR + 2*M.L.xR + 1*M.L.xZ <= 28)

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    M.pprint()

    with Solver('pao.pyomo.PCCG') as opt:
        results = opt.solve(M, tee=True, time_limit=10, quiet=False, maxit=100)

    print(results)
    M.display()




# 
# Example used in the PAO documentation
#   Simple bilevel
#
# Optimal solution: (x,y,z) = (6,4,2)
#
import numpy as np
import pyomo.environ as pe
import pao.mpr


def create():
    M = pao.mpr.LinearMultilevelProblem()

    U = M.add_upper(nxR=2)
    L = U.add_lower(nxR=1)

    U.x.lower_bounds = [2, np.NINF]
    U.x.upper_bounds = [6, np.PINF]
    L.x.lower_bounds = [0]
    L.x.upper_bounds = [np.PINF]

    U.c[U] = [1, 0]
    U.c[L] = [3]

    L.c[L] = [1]
    L.maximize = True

    U.equalities = True
    U.A[U] = [[1, 1]]
    U.b = [10]

    L.A[U] = [[ 1, 0],
              [-1, 0],
              [ 1, 0]]
    L.A[L] = [[ 1],
              [-4],
              [ 2]]
    L.b = [8, -8, 13]

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    opt = pao.Solver("pao.mpr.FA")
    opt.solve(M)
    print(M.U.x.values)
    print(M.U.LL.x.values)



# 
# Example used in the PAO documentation
#   Multiple lower-levels
#
# Optimal solution: (x,y,z) = (6,4,2)
#
import numpy as np
import pyomo.environ as pe
import pao.mpr


def create():
    M = pao.mpr.LinearMultilevelProblem()

    U = M.add_upper(nxR=2)
    L1 = U.add_lower(nxR=1)
    L2 = U.add_lower(nxR=1)

    U.x.lower_bounds = [2, np.NINF]
    U.x.upper_bounds = [6, np.PINF]
    U.c[U] = [1, 0]
    U.c[L1] = [3]
    U.c[L2] = [3]
    U.equalities = True
    U.A[U] = [[1, 1]]
    U.b = [10]

    L1.x.lower_bounds = [0]
    L1.x.upper_bounds = [np.PINF]
    L1.c[L1] = [1]
    L1.maximize = True
    L1.A[U] = [[ 1, 0],
               [-1, 0],
               [ 1, 0]]
    L1.A[L1] = [[ 1],
                [-4],
                [ 2]]
    L1.b = [8, -8, 13]

    L2.x.lower_bounds = [0]
    L2.x.upper_bounds = [np.PINF]
    L2.c[L2] = [1]
    L2.maximize = True
    L2.A[U] = [[0,  1],
               [0, -1],
               [0,  1]]
    L2.A[L2] = [[ 1],
                [-4],
                [ 2]]
    L2.b = [8, -8, 13]

    return M

if __name__ == "__main__":          #pragma: no cover
    M = create()
    opt = pao.Solver("pao.mpr.FA")
    opt.solve(M)
    for L in M.levels():
        print(L.name, L.x.values)



#
# bard511 example
# Using implicit index of lower level
# Using numpy/scipy data
#
import numpy as np
from scipy.sparse import coo_matrix
from pao.mpr import *


def create():
    M = LinearMultilevelProblem()

    U = M.add_upper(nxR=1)
    L = U.add_lower(nxR=1)

    U.x.lower_bounds = np.array([0])
    U.c[U] = np.array([1])
    U.c[L] = np.array([-4])

    L.x.lower_bounds = np.array([0])
    L.c[L] = np.array([1])

    L.A[U] = coo_matrix((np.array([-1, -2, 2, 3]),
                        (np.array([0, 1, 2, 3]),
                         np.array([0, 0, 0, 0]))))
    L.A[L] = coo_matrix((np.array([-1, 1, 1, -2]),
                        (np.array([0, 1, 2, 3]),
                         np.array([0, 0, 0, 0]))))
    L.b = np.array([-3, 0, 12, 4])

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    opt = Solver('pao.mpr.REG')
    res = opt.solve(M, tee=False)
    #M.print()
    print("PAO")
    print(res)



# Example 2.7 from
#
# Near-Optimal Robust Bilevel Optimization
#   M. Besancon, M. F. Anjos and L. Brotcorne
#   arXiv:1908.04040v5 (2019)
#
from pao.mpr import *


def create():
    M = LinearMultilevelProblem()

    U = M.add_upper(nxR=1)
    L = U.add_lower(nxR=1)

    # Variables
    U.x.lower_bounds = [2.5]    # Shifted solution here

    # Objectives
    U.c[U] = [1]
    U.c[L] = [-1]               # Adding pressure to maximize LL variable

    L.c[L] = [1]
    L.maximize = True

    # Constraints
    U.A[U] = [[-1/10]]
    U.A[L] = [[-1]]
    U.b = [-1]

    L.A[U] = [[-1/10]]
    L.A[L] = [[1]]
    L.b = [1]

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    opt = Solver('pao.mpr.FA')
    opt.solve(M, tee=True)
    M.print()



#
# bard511 example
# Using implicit index of lower level
# Using Python list data
#
from pao.mpr import *


def create():
    M = LinearMultilevelProblem()

    U = M.add_upper(nxR=1)
    L = U.add_lower(nxR=1)

    U.x.lower_bounds = [0]
    U.c[U] = [1]
    U.c[L] = [-4]

    L.x.lower_bounds = [0]
    L.c[L] = [1]

    L.A[U] = [[-1], [-2], [2], [3]]
    L.A[L] = [[-1], [1], [1], [-2]]

    L.b = [-3, 0, 12, 4]

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    opt = Solver('pao.mpr.FA')
    opt.solve(M)
    M.print()



#
# Toy Example 2 from 
#
# "A projection-based reformulation and decomposition algorithm for global optimization 
#  of a class of mixed integer bilevel linear programs" by Dajun Yue, Jiyao Gao, Bo Zeng, Fengqi You
#
import pyomo.environ as pe
from pao.pyomo import *


def create():
    M = pe.ConcreteModel()

    M.xZ = pe.Var(bounds=(0,None), within=pe.Integers)

    M.L = SubModel(fixed=(M.xZ))
    M.L.xZ = pe.Var(bounds=(0,None), within=pe.Integers)

    M.o = pe.Objective(expr=-M.xZ - 2*M.L.xZ, sense=pe.minimize)
    M.c1 = pe.Constraint(expr=-2*M.xZ + 3*M.L.xZ <= 12)
    M.c2 = pe.Constraint(expr=M.xZ + M.L.xZ <= 14)

    M.L.o = pe.Objective(expr=M.L.xZ, sense=pe.maximize)
    M.L.c1 = pe.Constraint(expr=-3*M.xZ + M.L.xZ <= -3)
    M.L.c2 = pe.Constraint(expr=3*M.xZ + M.L.xZ <= 30)

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    M.pprint()



#
# Example from
# Moore, J. and J. Bard 1990.
# The mixed integer linear bilevel programming problem.
# Operations Research 38(5), 911–921.
#
import pyomo.environ as pe
from pao.pyomo import *


def create():
    M = pe.ConcreteModel()

    M.xZ = pe.Var(within=pe.Integers)

    M.L = SubModel(fixed=(M.xZ))
    M.L.xZ = pe.Var(within=pe.Integers)

    M.o = pe.Objective(expr=-M.xZ - 10*M.L.xZ)

    M.L.o = pe.Objective(expr=M.L.xZ)
    M.L.c1 = pe.Constraint(expr= -25*M.xZ + 20*M.L.xZ <=  30)
    M.L.c2 = pe.Constraint(expr=     M.xZ +  2*M.L.xZ <=  10)
    M.L.c3 = pe.Constraint(expr=   2*M.xZ -    M.L.xZ <=  15)
    M.L.c4 = pe.Constraint(expr=  -2*M.xZ - 10*M.L.xZ <= -15)

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    M.pprint()



import math
import weakref
import copy
import pprint
import collections.abc
from scipy.sparse import csr_matrix, dok_matrix
import numpy as np
from pyutilib.misc import Bunch


def _equal_nparray(Ux, U_coef, Lx, L_coef):
    if Ux is None and Lx is None:
        return True
    if Ux is None or Lx is None:
        return False
    for i in range(Ux.size):
        if math.fabs(Ux[i]*U_coef + Lx[i]*L_coef) > 1e-16:
            return False
    return True

def _update_matrix(*, A, old, new, update_columns=True):
    A = A.todok()
    if not update_columns:
        A = A.transpose()

    A_ = dok_matrix((A.shape[0], new.nxR+new.nxZ+new.nxB), dtype=np.float64)
    for ndx in A.keys():
        i,j = ndx
        if j < old.nxR:
            if j<new.nxR:
                A_[i,j] = A[i,j]
        elif j < old.nxR+old.nxZ:
            j_ = j-old.nxR
            if j_<new.nxZ:
                A_[i,j_+new.nxR] = A[i,j]
        else:
            j_ = j-old.nxR-old.nxZ
            if j_<new.nxB:
                A_[i,j_+new.nxR+new.nxZ] = A[i,j]

    if not update_columns:
        A_ = A_.transpose()
    return A_.tocsr()


class SimplifiedList(collections.abc.MutableSequence):
    """
    This is a normal list class, except if the user asks to
    get attributes, then they will be collected from the
    first list element.  However, in that case, an error
    is generated if there are more than one list elements.
    """

    def __init__(self):
        self._data = []

    def clone(self, parent=None, clone_fn=None):
        ans = SimplifiedList()
        ans._data = [ val.clone(parent=parent, clone_fn=clone_fn) for val in self._data ]
        return ans

    def append(self, val):
        self._data.append(val)

    def insert(self, i, val):
        self._data.insert(i, val)

    def __iter__(self):
        for v in self._data:
            yield v

    def __getitem__(self, i):
        if i >= len(self._data):
            raise IndexError
        return self._data[i]

    def __setitem__(self, i, val):
        self._data[i] = val

    def __delitem__(self, i):
        del self._data[i]

    def __len__(self):
        return len(self._data)

    def __getattr__(self, name):
        if not name.startswith('_'):
            assert (len(self._data) == 1), "Getting attributes of a simplified list, which has %d elements" % len(self._data)
            return getattr(self._data[0], name)
        return super().__getattr__(name)

    def __setattr__(self, name, val):
        if not name.startswith('_'):
            assert (len(self._data) == 1), "Setting attributes of a simplified list, which has %d elements" % len(self._data)
            return setattr(self._data[0], name, val)
        return super().__setattr__(name, val)


class LevelVariable(object):

    def __init__(self, nxR=0, nxZ=0, nxB=0, lb=None, ub=None):
        self.nxR = nxR
        self.nxZ = nxZ
        self.nxB = nxB
        self.num = nxR+nxZ+nxB
        self.values = [None]*self.num
        self.pyvar = [None]*self.num
        if lb is None:
            self.lower_bounds = np.array([np.NINF]*self.num, dtype=np.float64)
        else:
            self.lower_bounds = lb
        if ub is None:
            self.upper_bounds = np.array([np.PINF]*self.num, dtype=np.float64)
        else:
            self.upper_bounds = ub
        for i in range(nxB):
            self.lower_bounds[i+nxR+nxZ] = 0
            self.upper_bounds[i+nxR+nxZ] = 1

    def clone(self):
        ans = LevelVariable(self.nxR, self.nxZ, self.nxB, np.copy(self.lower_bounds), np.copy(self.upper_bounds))
        ans.values = copy.copy(self.values)
        ans.pyvar = self.pyvar
        return ans

    def __len__(self):
        return self.num

    def __iter__(self):
        for i in range(self.num):
            yield i

    def _resize(self, nxR, nxZ, nxB, lb=np.NINF, ub=np.PINF):
        if nxR == self.nxR and nxZ == self.nxR and nxB == self.nxB:     # pragma: nocover
            return

        num = nxR+nxZ+nxB
        curr_num = self.num
        self.num = num
        self.values = [None]*num
        self.pyvar = [None]*num

        tlower = self.lower_bounds
        tupper = self.upper_bounds
        self.lower_bounds = np.array([lb]*num, dtype=np.float64)
        self.upper_bounds = np.array([ub]*num, dtype=np.float64)

        for i in range(min(nxR,self.nxR)):
            self.lower_bounds[i] = tlower[i]
            self.upper_bounds[i] = tupper[i]
        for i in range(min(nxZ,self.nxZ)):
            self.lower_bounds[i+nxR] = tlower[i+self.nxR]
            self.upper_bounds[i+nxR] = tupper[i+self.nxR]
        for i in range(min(nxB,self.nxB)):
            self.lower_bounds[i+nxR+nxZ] = tlower[i+self.nxR+self.nxZ]
            self.upper_bounds[i+nxR+nxZ] = tupper[i+self.nxR+self.nxZ]

        self.nxR = nxR
        self.nxZ = nxZ
        self.nxB = nxB

    def print(self):                  # pragma: no cover
        print("  nxR: "+str(self.nxR))
        print("  nxZ: "+str(self.nxZ))
        print("  nxB: "+str(self.nxB))
        print("  lower bounds: "+str(self.lower_bounds))
        print("  upper bounds: "+str(self.upper_bounds))
        print("  nonzero values:")
        for i,v in enumerate(self.values):
            if v is not None and v != 0:
                if type(v) is int:
                    print("    %d: %d" % (i, v))
                else:
                    print("    %d: %f" % (i, v))

    def __setattr__(self, name, value):
        if name == 'lower_bounds':
            assert (value is not None), "Cannot specify null lower bounds array"
            # Add this check in the model checks
            assert (len(value) == self.num), "The variable has length %s but specifying a lower bounds with length %s" % (str(self.num), str(len(value)))
            if type(value) is list:
                value = np.array(value, dtype=np.float64)
            super().__setattr__(name, value)
        elif name == 'upper_bounds':
            assert (value is not None), "Cannot specify null upper bounds array"
            # Add this check in the model checks
            assert (len(value) == self.num), "The variable has length %s but specifying a upper bounds with length %s" % (str(self.num), str(len(value)))
            if type(value) is list:
                value = np.array(value, dtype=np.float64)
            super().__setattr__(name, value)
        else:
            super().__setattr__(name, value)


class LevelValues(object):

    def __init__(self, matrix_list=False, matrix=False, x=None):
        self._matrix = matrix
        self._matrix_list = matrix_list
        self.set_values(x)

    def clone(self):
        ans = LevelValues(matrix_list=self._matrix_list, matrix=self._matrix)
        if self._matrix:
            if self.x is not None:
                ans.x = self.x.copy()
        elif self._matrix_list:
            ans.x = [None if m is None else np.copy(m) for m in self.x]
        else:
            if self.x is not None:
                ans.x = np.copy(self.x)
        return ans

    def set_values(self, x=None):
        if type(x) is list:
            if self._matrix:                
                x = csr_matrix( x )
            elif self._matrix_list:                
                x = [csr_matrix(m) if type(m) is list else m for m in x]
            else:
                x = np.array(x, dtype=np.float64)
        elif type(x) is tuple:
            if self._matrix:
                nrows,ncols = x[0]
                xnew = dok_matrix((nrows,ncols))
                for key,value in x[1].items():
                    xnew[key] = value
                x = xnew
            elif self._matrix_list:
                ncon,nrows,ncols = x[0]
                con = {}
                for key,value in x[1].items():
                    i,j,k = key
                    if i not in con:
                        con[i] = dok_matrix((nrows,ncols))
                    con[i][j,k] = value
                xnew = []
                for i in range(ncon):
                    if i in con:
                        xnew.append( csr_matrix(con[i]) )
                    else:
                        xnew.append( None )
                x = xnew
                
        super().__setattr__('x', x)

    def __setattr__(self, name, value):
        if name == 'x':
            self.set_values(value)
        else:
            super().__setattr__(name, value)

    def print(self, prefix):                        # pragma: no cover
        self._print_value(self.x, prefix)

    def __len__(self):
        n = 0
        if self._matrix:
            if self.x is not None:
                n = max(n, self.x.shape[0])
        elif self._matrix_list:
            if self.x is not None:
                n += sum(0 if m is None else m.size for m in self.x)
        else:
            if self.x is not None:
                n += self.x.size
        return n

    def _print_value(self, value, name):            # pragma: no cover
        if value is not None:
            if self._matrix:
                print("    %s:" % name)
                if value.size:
                    print("        shape: %d %d" % (value.shape[0], value.shape[1]))
                print("        nonzeros:")
                for row in value.toarray().tolist():
                    print("        ",row)
            elif self._matrix_list:
                print("    %s:" % name)
                for i,m in enumerate(value):
                    print("    %d:" % i)
                    if m is None:
                        print("        None")
                        continue
                    print("        shape: %d %d" % (m.shape[0], m.shape[1]))
                    print("        nonzeros:")
                    for row in str(m).split('\n'):
                        print("        "+row)
            else:
                print("    %s:" % name, value)


class LevelValueWrapper1(object):

    def __init__(self, prefix, matrix=False):
        setattr(self, '_matrix', matrix)
        setattr(self, '_values', {})
        setattr(self, '_prefix', prefix)

    def __len__(self):
        _values = getattr(self, '_values')
        return len(_values)

    def __iter__(self):
        yield from self._values.keys()

    def __getattr__(self, name):
        if name.startswith('_'):                # pragma: no cover
            return super().__getattr__(name)
        else:
            raise AttributeError("No attributes in this object")

    def __getitem__(self, lvl):
        if type(lvl) is int:
            i = lvl
        else:
            i = lvl.id
        _values = self._values
        retval = _values.get(i, None)
        if retval is None:
            return None
        return retval.x

    def __setitem__(self, lvl, value):
        if type(lvl) is int:
            i = lvl
        else:
            i = lvl.id
        _values = self._values
        if value is None:
            if i in _values:
                del _values[i]
        else:
            _values[i] = LevelValues(matrix=self._matrix, x=value)

    def clone(self):
        ans = LevelValueWrapper1(self._prefix, matrix=self._matrix)
        for name in self._values:
            ans._values[name] = self._values[name].clone()
        return ans

    def print(self, names):               # pragma: no cover
        _values = self._values
        first = True
        for i,name in names:
            v = _values.get(i,None)
            if v is None:
                continue
            if len(v) > 0:
                if first:
                    print("  "+self._prefix+":")
                    first = False
                v.print(name)


class LevelValueWrapper2(object):

    def __init__(self, prefix, matrix=True):
        setattr(self, '_matrix', matrix)
        setattr(self, '_values', {})
        setattr(self, '_prefix', prefix)

    def __len__(self):
        _values = getattr(self, '_values')
        return len(_values)

    def __iter__(self):
        yield from self._values.keys()

    def __getattr__(self, name):
        if name.startswith('_'):                # pragma: no cover
            return super().__getattr__(name)
        else:
            raise AttributeError("No attributes in this object")

    def __getitem__(self, lvls):
        lvl1, lvl2 = lvls
        if type(lvl1) is int:
            i = lvl1
        else:
            i = lvl1.id
        if type(lvl2) is int:
            j = lvl2
        else:
            j = lvl2.id
        _values = self._values
        retval = _values.get((i,j), None)
        if retval is None:
            return None
        return retval.x

    def __setitem__(self, lvls, value):
        lvl1, lvl2 = lvls
        if type(lvl1) is int:
            i = lvl1
        else:
            i = lvl1.id
        if type(lvl2) is int:
            j = lvl2
        else:
            j = lvl2.id
        assert (i<=j), "Require quadratic terms to be indexed with upper-level variables first"
        if not type(lvl1) is int and not type(lvl2) is int:
            assert (lvl1.id in [L.id for L in lvl2.levels()]), "Require quadratic terms to be in the same subproblem tree"
        if value is None:
            if (i,j) in self._values:
                del self._values[i,j]
        else:
            self._values[i,j] = LevelValues(matrix=self._matrix, matrix_list=not self._matrix, x=value)

    def clone(self):
        ans = LevelValueWrapper2(self._prefix, matrix=self._matrix)
        for name in self._values:
            ans._values[name] = self._values[name].clone()
        return ans

    def print(self, names):               # pragma: no cover
        _values = self._values
        first = True
        for i,name in names:
            v = _values.get(i,None)
            if v is None:
                continue
            if len(v) > 0:
                if first:
                    print("  "+self._prefix+":")
                    first = False
                v.print(name)


class LinearLevelRepn(object):

    _counter = 0

    def __init__(self, nxR, nxZ, nxB, id=None):
        if id is None:
            self.id = LinearLevelRepn._counter
            LinearLevelRepn._counter += 1
        else:
            self.id = id

        self.x = LevelVariable(nxR, nxZ, nxB)    # variables at this level
        self.c = LevelValueWrapper1("c") # objective coefficients at this level
        self.A = LevelValueWrapper1("A",
                        matrix=True)    # constraint matrices at this level
        self.b = np.ndarray(0, dtype=np.float64)          # RHS of the constraints
        self._minimize = True           # sense of the objective at this level
        self._inequalities = True       # If True, the constraints are inequalities
        self.d = 0                      # constant in objective at this level

        self.LL = SimplifiedList()      # lower levels
        self.UL = lambda: None          # "empty weakref" to upper level

        self.name = None                # a string descriptor for this level

    def _add_lower(self, tmp, nxR=0, nxZ=0, nxB=0, name=None, id=None):
        if name is None:
            tmp.name = self.name + ".LL[%d]" % len(self.LL)
        else:
            tmp.name = name
        tmp.UL = weakref.ref(self)
        self.LL.append(tmp)
        return tmp

    def add_lower(self, *, nxR=0, nxZ=0, nxB=0, name=None, id=None):
        return self._add_lower(LinearLevelRepn(nxR, nxZ, nxB, id=id), nxR=nxR, nxZ=nxZ, name=name, id=id)

    @property
    def inequalities(self):
        return self._inequalities

    @inequalities.setter
    def inequalities(self, val):
        self._inequalities = val

    @property
    def equalities(self):
        return not self._inequalities

    @equalities.setter
    def equalities(self, val):
        self._inequalities = not val

    @property
    def minimize(self):
        return self._minimize

    @minimize.setter
    def minimize(self, val):
        self._minimize = val

    @property
    def maximize(self):
        return not self._minimize

    @maximize.setter
    def maximize(self, val):
        self._minimize = not val

    def resize(self, *, nxR=0, nxZ=0, nxB=0, lb=np.NINF, ub=np.PINF):
        old = Bunch(nxR=self.x.nxR, nxZ=self.x.nxZ, nxB=self.x.nxB)
        new = Bunch(nxR=nxR, nxZ=nxZ, nxB=nxB)
        self.x._resize(nxR=nxR, nxZ=nxZ, nxB=nxB, lb=lb, ub=ub)
        #
        # Walk the tree, updating c[self] and A[self]
        #
        for L in self.levels():
            L._update(level=self, new=new, old=old)

    def _update(self, *, level, new, old):
        #
        # Update 'c'
        #
        c = self.c[level]
        if c is not None:
            c_ = np.zeros(new.nxR+new.nxZ+new.nxB)          # RHS of the constraints
            for i in range(min(new.nxR,old.nxR)):
                c_[i] = c[i]
            for i in range(min(new.nxZ,old.nxZ)):
                c_[i+new.nxR] = c[i+old.nxR]
            for i in range(min(new.nxB,old.nxB)):
                c_[i+new.nxR+new.nxZ] = c[i+old.nxR+old.nxZ]
            self.c[level] = c_
        #
        # Update 'A'
        #
        A = self.A[level]
        if A is not None:
            self.A[level] = _update_matrix(A=A, old=old, new=new)

    #
    # Iterate over the sublevels and parents in DFS order
    #
    def levels(self, parents=True):
        for X in self._sublevels():
            yield X
        if parents:
            X = self.UL()
            while X is not None:
                yield X
                X = X.UL()

    #
    # Iterate over sublevels in DFS order
    #
    def _sublevels(self):
        yield self
        for L in self.LL:
            for X in L._sublevels():
                yield X

    @staticmethod
    def _clone_level(self, parent=None, data=[], ans=None):
        if ans is None:
            ans = LinearLevelRepn(0,0,0)
        ans.x = self.x.clone()
        ans.c = self.c.clone()
        ans.A = self.A.clone()
        ans.b = np.copy(self.b)
        ans.minimize = self.minimize
        ans.inequalities = self.inequalities
        ans.d = self.d
        if parent is None:
            ans.UL = lambda: None # "empty weakref"
        else:
            ans.UL = weakref.ref(parent)
        # TODO - Should we allow users to annotate these objects with other data?
        for attr in dir(self):
            if attr in data:
                continue
            if attr in ['clone', 'print', 'resize', 'add_lower', 'levels']:    # methods
                continue
            if attr.startswith('_'):
                continue
            setattr(ans, attr, copy.copy(getattr(self, attr)))
        return ans

    def clone(self, parent=None, clone_fn=None):
        if clone_fn is None:
            clone_fn = LinearLevelRepn._clone_level
        ans = clone_fn(self, parent=parent, data=['x', 'c', 'A', 'b', 'minimize', 'inequalities', 'equalities', 'd', 'LL', 'UL'])
        ans.LL = self.LL.clone(parent=ans, clone_fn=clone_fn)
        return ans

    def print(self, names):       # pragma: no cover
        print("")
        print("## Level: "+self.name)
        print("")

        print("Variables:")
        self.x.print()

        print("\nObjective:")
        if self.minimize:
            print("  Minimize:")
        else:
            print("  Maximize:")
        self.c.print(names)
        print("  d:",self.d)

        if self.b.size > 0:
            print("\nConstraints: ")
            self.A.print(names)
            if self.inequalities:
                print("  <=", self.b)
            else:
                print("  ==", self.b)
        #
        # Recurse
        #
        for L in self.LL:
            L.print(names)

    def __setattr__(self, name, value):
        if name == 'b' and value is not None:
            value = np.array(value, dtype=np.float64)
            super().__setattr__(name, value)
        else:
            super().__setattr__(name, value)

    def check(self):
        #
        # Size of 'c'
        #
        for X in self.levels():
            assert ((self.c[X] is None) or (self.c[X].size == len(X.x)) or (self.c[X].size == 0)), "Incompatible specification of coefficients for %s.c[%s]: %d != %d" % (self.name, X.name, self.c[X].size, len(X.x))
        #
        # Ncols of 'A'
        #
        for X in self.levels():
            assert ((self.A[X] is None) or (self.A[X].shape[1] == len(X.x))), "Incompatible specification of %s.A[%s] and %s.x (%d != %d)" % (self.name, X.name, X.name, self.A[X].shape[1], len(X.x))
        #
        # Nrows of 'A'
        #
        if self.b is None:
            for X in self.levels():
                assert (self.A[X] is None), "Incompatible specification of %s.b and %s.A[%s]" % (self.name, self.name, X.name)
        else:
            nr = self.b.size
            for X in self.levels():
                if self.A[X] is not None:
                    assert (nr == self.A[X].shape[0]), "Incompatible specification of %s.b and %s.A[%s] (%d != %d)" % (self.name, self.name, X.name, nr, self.A[X].shape[0])


class QuadraticLevelRepn(LinearLevelRepn):

    def __init__(self, nxR, nxZ, nxB, id=None):
        LinearLevelRepn.__init__(self, nxR, nxZ, nxB, id)
        self.P = LevelValueWrapper2("P", matrix=True)        # objective quadratic terms
        self.Q = LevelValueWrapper2("Q", matrix=False)       # constraint quadratic terms

    def add_lower(self, *, nxR=0, nxZ=0, nxB=0, name=None, id=None):
        return self._add_lower(QuadraticLevelRepn(nxR, nxZ, nxB, id=id), nxR=nxR, nxZ=nxZ, name=name, id=id)

    def _update(self, *, level, new, old):
        #
        # Update 'c' and 'A'
        #
        LinearLevelRepn._update(self, level=level, new=new, old=old)
        #
        # Update 'P'
        #
        for L1,L2 in self.P:
            if L1 == level.id or L2 == level.id:
                self.P[L1,L2] = _update_matrix(A=self.P[L1,L2], old=old, new=new, update_columns=L2==level.id)
        #
        # Update 'Q'
        #
        for L1,L2 in self.Q:
            if L1 == level.id or L2 == level.id:
                Q = self.Q[L1,L2]
                self.Q[L1,L2] = [None if m is None else _update_matrix(A=m, old=old, new=new, update_columns=L2==level.id) for m in Q]

    @staticmethod
    def _clone_level(self, parent, data, ans=None):
        if ans is None:
            ans = QuadraticLevelRepn(0,0,0)
        LinearLevelRepn._clone_level(self, parent, data, ans=ans)
        ans.P = self.P.clone()
        ans.Q = self.Q.clone()
        return ans
        
    def clone(self, parent=None, clone_fn=None):
        if clone_fn is None:
            clone_fn = QuadraticLevelRepn._clone_level
        ans = clone_fn(self, parent=parent, data=['x', 'c', 'A', 'b', 'minimize', 'maximize', 'inequalities', 'equalities', 'd', 'LL', 'UL', 'P', 'Q'])
        ans.LL = self.LL.clone(parent=ans, clone_fn=clone_fn)
        return ans

    def print(self, names):       # pragma: no cover
        print("")
        print("## Level: "+self.name)
        print("")

        print("Variables:")
        self.x.print()

        print("\nObjective:")
        if self.minimize:
            print("  Minimize:")
        else:
            print("  Maximize:")
        self.c.print(names)
        self.P.print(names)
        print("  d:",self.d)

        if self.b.size > 0:
            print("\nConstraints: ")
            self.A.print(names)
            self.Q.print(names)
            if self.inequalities:
                print("  <=", self.b)
            else:
                print("  ==", self.b)
        #
        # Recurse
        #
        for L in self.LL:
            L.print(names)

    def check(self):
        LinearLevelRepn.check(self)
        #
        # Ncols of 'P'
        #
        for X in self.levels():
            if X.id <= self.id:
                assert ((self.P[X,self] is None) or (self.P[X,self].shape[1] == len(self.x))), "Incompatible specification of columns for %s.P[%s,%s] and %s.x (%d != %d)" % (self.name, X.name, self.name, self.name, self.P[X,self].shape[1], len(self.x))


class LinearMultilevelProblem(object):
    """
    ::

      For bilevel problems, let:

        U   = LinearMultilevelProblem.U
        L   = U.L
        x   = [U.x, L.x]'       # dense column vector
        U.c = [U.c[U], U.c[L]]' # dense column vector
        L.c = [L.c[U], L.c[L]]' # dense column vector
        U.A = [U.A[U], U.A[L]]  # sparse matrix
        L.A = [L.A[U], L.A[L]]  # sparse matrix

      Then we have:

        min_{U.x}   U.c' * x + U.d
        s.t.        U.A  * x       <= U.b                  # Or ==

                where L.x satisifies

                    min_{L.x}   L.c' * x + L.d
                    s.t.        L.A  * x       <= L.b      # Or ==
    """

    def __init__(self, name=None):
        self.name = name
        self.U = None

    def add_upper(self, *, nxR=0, nxZ=0, nxB=0, name=None, id=None):
        assert (self.U is None), "Cannot create a second upper-level in a LinearMultilevelProblem"
        self.U = LinearLevelRepn(nxR, nxZ, nxB, id=id)
        if name is None:
            self.U.name = "U"
        else:
            self.U.name = name
        return self.U

    def levels(self):
        yield from self.U._sublevels()

    def clone(self, clone_fn=None):
        ans = LinearMultilevelProblem()
        ans.name = self.name
        ans.U = self.U.clone(clone_fn=clone_fn)
        return ans

    def print(self):                            # pragma: no cover
        nL = len(self.U.LL)
        if self.name:
            print("# LinearMultilevelProblem: "+self.name)
        else:
            print("# LinearMultilevelProblem: unknown")

        names = [(L.id,L.name) for L in self.levels()]
        self.U.print(names)

    def check(self):                    # pragma: no cover
        for L in self.levels():
            L.check()

    def check_opposite_objectives(self, U, L):
        if id(U.c) == id(L.c) and L.minimize ^ U.minimize:
            return True
        U_coef = 1 if U.minimize else -1
        L_coef = 1 if L.minimize else -1
        if not _equal_nparray(U.c[U], U_coef, L.c[U], L_coef):
            return False
        if not _equal_nparray(U.c[L], U_coef, L.c[L], L_coef):
            return False
        return True


class QuadraticMultilevelProblem(object):
    """
    ::
 
      For bilevel problems, let:
    
        U   = QuadraticMultilevelProblem.U
        L   = U.L
        x   = [U.x, L.x]'           # dense column vector
        U.c = [U.c[U], U.c[L]]'     # dense column vector
        U.P = [U.P[U,U], U.P[U,L]]  # sparse matrix
              [0,        U.P[L,L]]
        U.A = [U.A[U], U.A[L]]      # sparse matrix
        U.Q = [U.Q[U,U], U.Q[U,L]]  # sparse matrix
              [0,        U.Q[L,L]]
        L.c = [L.c[U], L.c[L]]'     # dense column vector
        L.P = [L.P[U,U], L.P[U,L]]  # sparse matrix
              [0,        L.P[L,L]]
        L.A = [L.A[U], L.A[L]]      # sparse matrix
        L.Q = [L.Q[U,U], L.Q[U,L]]  # sparse matrix
              [0,        L.Q[L,L]]
    
      Then we have:
    
        min_{U.x}   U.c' * x + x' * U.P * x + U.d
        s.t.        U.A  * x + x' * U.Q * x       <= U.b                 # Or ==

                where L.x satisifies

                    min_{L.x}   L.c' * x + x' * L.P * x + L.d
                    s.t.        L.A  * x + x' * L.Q * x       <= L.b     # Or ==
    """

    def __init__(self, name=None, bilinear=False):
        self.bilinear = bilinear
        self.name = name
        self.U = None

    def add_upper(self, *, nxR=0, nxZ=0, nxB=0, name=None, id=None):
        assert (self.U is None), "Cannot create a second upper-level in a QuadraticMultilevelProblem"
        self.U = QuadraticLevelRepn(nxR, nxZ, nxB, id=id)
        if name is None:
            self.U.name = "U"
        else:
            self.U.name = name
        return self.U

    def levels(self):
        yield from self.U._sublevels()

    def clone(self, clone_fn=None):
        ans = QuadraticMultilevelProblem()
        ans.name = self.name
        ans.U = self.U.clone(clone_fn=clone_fn)
        return ans

    def print(self):                            # pragma: no cover
        nL = len(self.U.LL)
        if self.name:
            print("# QuadraticMultilevelProblem: "+self.name)
        else:
            print("# QuadraticMultilevelProblem: unknown")

        names = [(L.id,L.name) for L in self.levels()]
        for L in self.levels():
            for X in L.levels():
                if L.id <= X.id:
                    names.append( ((L.id,X.id), L.name+","+X.name) )
        self.U.print(names)

    def check(self):                    # pragma: no cover
        for L in self.levels():
            L.check()

    def check_opposite_objectives(self, U, L):
        if id(U.c) == id(L.c) and L.minimize ^ U.minimize:
            return True
        U_coef = 1 if U.minimize else -1
        L_coef = 1 if L.minimize else -1
        if not _equal_nparray(U.c[U], U_coef, L.c[U], L_coef):
            return False
        if not _equal_nparray(U.c[L], U_coef, L.c[L], L_coef):
            return False
        return True


if __name__ == "__main__":              # pragma: no cover
    prob = LinearMultilevelProblem()
    U = prob.add_upper(nxR=3,nxZ=2,nxB=1)
    U.x.upper_bounds = np.array([1.5, 2.4, 3.1, np.PINF, np.PINF, 1])
    L = U.add_lower(nxR=1,nxZ=2,nxB=3)
    L.x.lower_bounds = np.array([np.NINF, 1, -2, 0, 0, 0])
    prob.print()




#
# A solver for linear bilevel programs using big-M 
# relaxations discussed by Fortuny-Amat and McCarl, 1981.
#
import time
import numpy as np
import pyutilib
import pyomo.environ as pe
import pyomo.opt
from pyomo.common.config import ConfigBlock, ConfigValue
from pyomo.mpec import ComplementarityList, complements

import pao.common
from ..solver import Solver, LinearMultilevelSolverBase, LinearMultilevelResults
from ..repn import LinearMultilevelProblem
from ..convert_repn import convert_to_standard_form
from . import pyomo_util
from .reg import create_model_replacing_LL_with_kkt


@Solver.register(
        name='pao.mpr.FA',
        doc='PAO solver for Multilevel Problem Representations that define linear bilevel problems.  Solver uses big-M relaxations discussed by Fortuny-Amat and McCarl (1981).')

class LinearMultilevelSolver_FA(LinearMultilevelSolverBase):
    """
    PAO FA solver for linear MPRs: pao.mpr.FA

    This solver replaces lower-level problems using the KKT conditions and
    calls a MIP solver to solve the reformulated problem.
    """
    config = LinearMultilevelSolverBase.config()
    config.declare('mip_solver', ConfigValue(
        default='glpk',
        description="The MIP solver used by FA.  (default is glpk)"
        ))
    #config.declare('mip_options', ConfigValue(
    #    default=None,
    #    description="A dictionary that defines the solver options for the MIP solver.  (default is None)"))
    config.declare('bigm', ConfigValue(
        default=100000,
        domain=float,
        description="The big-M value used to enforce complementarity conditions.  (default is 1e5)"
        ))

    def __init__(self, **kwds):
        super().__init__(name='pao.mpr.FA')

    def check_model(self, model):
        #
        # Confirm that the LinearMultilevelProblem is well-formed
        #
        assert (type(model) is LinearMultilevelProblem), "Solver '%s' can only solve a linear multilevel problem" % self.name
        model.check()
        #
        # Confirm that this is a bilevel problem
        #
        for L in model.U.LL:
            assert (len(L.LL) == 0), "Can only solve bilevel problems"
        #
        # No binary or integer lower level variables
        #
        for L in model.U.LL:
            assert (L.x.nxZ == 0), "Cannot use solver %s with model with integer lower-level variables" % self.name
            assert (L.x.nxB == 0), "Cannot use solver %s with model with binary lower-level variables" % self.name

    def solve(self, model, **options):
        #
        # Error checks
        #
        self.check_model(model)
        #
        # Process keyword options
        #
        self._update_config(options)
        #
        # Start clock
        #
        start_time = time.time()

        self.standard_form, soln_manager = convert_to_standard_form(model, inequalities=False)

        M = self._create_pyomo_model(self.standard_form, self.config.bigm)
        #
        # Solve the Pyomo model the specified solver
        #
        results = LinearMultilevelResults(solution_manager=soln_manager)
        if isinstance(self.config.mip_solver, str):
            opt = pe.SolverFactory(self.config.mip_solver)
        else:
            opt = self.config.mip_solver

        #if self.config.mip_options is not None:
        #    opt.options.update(self.config.mip_options)
        pyomo_results = opt.solve(M, tee=self.config.tee, 
                                     load_solutions=self.config.load_solutions)
        pyomo.opt.check_optimal_termination(pyomo_results)

        self._initialize_results(results, pyomo_results, M)
        results.solver.rc = getattr(opt, '_rc', None)

        if self.config.load_solutions:
            # Load results from the Pyomo model to the LinearMultilevelProblem
            results.copy_solution(From=M, To=model)
        else:
            # Load results from the Pyomo model to the Results
            results.load_from(pyomo_results)

        #self._debug(M)
        #results.solver.log = getattr(opt, '_log', None)

        results.solver.wallclock_time = time.time() - start_time
        return results

    def _initialize_results(self, results, pyomo_results, M):
        #
        # SOLVER
        #
        solv = results.solver
        solv.name = self.config.mip_solver
        solv.termination_condition = pyomo_util.pyomo2pao_termination_condition(pyomo_results.solver.termination_condition)
        if hasattr(pyomo_results.solver, 'time'):
            solv.solver_time = pyomo_results.solver.time
        if self.config.load_solutions:
            solv.best_feasible_objective = pe.value(M.o)
        #
        # PROBLEM - Maybe this should be the summary of the BLP itself?
        #
        prob = results.problem
        prob.name = M.name
        #prob.number_of_objectives = pyomo_results.problem.Number_of_objectives
        #prob.number_of_constraints = pyomo_results.problem.Number_of_constraints
        #prob.number_of_variables = pyomo_results.problem.Number_of_variables
        #prob.number_of_nonzeros = pyomo_results.problem.Number_of_nonzeros
        #prob.lower_bound = pyomo_results.problem.Lower_bound
        #prob.upper_bound = pyomo_results.problem.Upper_bound
        #prob.sense = 'minimize'
        return results

    def _create_pyomo_model(self, repn, bigM):
        M = create_model_replacing_LL_with_kkt(repn)
        #
        # Transform the problem to a MIP
        #
        # TODO - directly create the bigM relaxation here.  Applying
        # Pyomo transformations in sequence creates a model object that is
        # difficult to interpret.
        #
        xfrm = pe.TransformationFactory('mpec.simple_disjunction')
        xfrm.apply_to(M)
        xfrm = pe.TransformationFactory('gdp.bigm')
        xfrm.apply_to(M, bigM=bigM)

        return M

    def _debug(self, M):    # pragma: no cover
        for j in M.U.xR:
            print("U",j,pe.value(M.U.xR[j]))
        for j in M.L.xR:
            print("L",j,pe.value(M.L.xR[j]))
        for j in M.kkt.lam:
            print("lam",j,pe.value(M.kkt.lam[j]))
        for j in M.kkt.nu:
            print("nu",j,pe.value(M.kkt.nu[j]))


pao.common.SolverAPI._generate_solve_docstring(LinearMultilevelSolver_FA)




'''
Implementation of the algorithm presented in "A projection-based reformulation
and decomposition algorithm for global optimization of a class of mixed integer 
bilevel linear programs" by Dajun Yue, Jiyao Gao, Bo Zeng, Fengqi You

Implemented May-August 2019 by She'ifa Punla-Green at Sandia National Labs

This algorithm seeks to solve the following bilevel MILP:
    min cR*xu + cZ*yu + dR*xl0 + dZ*yl0 
    s.t. AR*xu + AZ*yu + BR*xl0 + BZ* yl0 <= r
     (xl0,yl0) in argmax {wR*xl+wZ*yl: PR*xl+PZ*yl<=s-QR*xu-QZ*yu}
'''
import time

from pyomo.environ import *
from pyomo.gdp import *
from pyomo.mpec import *

from . import pyomo_util


infinity = float('inf')

offset=0

def mat2dict(m, start, stop):
    if m is None:
        return {}
    cx = m.tocoo()    
    return {(i+offset,j+offset-start):float(v) for i,j,v in zip(cx.row, cx.col, cx.data) if j>=start and j<stop}

def array2dict(m, start=0, stop=None):
    if m is None:
        return {}
    if stop is None:
        stop = len(m)
    return {i+offset-start:float(m[i]) for i in range(len(m)) if i>=start and i<stop}


def create_pyomo_model(mpr, M):
    '''
    Parameter Import

    Notation:
    m is for upper level
    n is for lower level
    x is continuous (R)
    y is discrete (Z)

    coefficient vectors and matrices should be dictionaries (with tuples for matrices)
    variable size should be floats
    '''
    U = mpr.U
    L = mpr.U.LL[0]

    mU = len(U.b)
    mR = U.x.nxR
    mZ = U.x.nxZ
    nL = len(L.b)
    nR = L.x.nxR
    nZ = L.x.nxZ

    AR = mat2dict(U.A[U], 0,       U.x.nxR)
    AZ = mat2dict(U.A[U], U.x.nxR, U.x.nxR+U.x.nxZ)
    BR = mat2dict(U.A[L], 0,       L.x.nxR)
    BZ = mat2dict(U.A[L], L.x.nxR, L.x.nxR+U.x.nxZ)
    r  = array2dict(U.b)
    cR = array2dict(U.c[U], 0,       U.x.nxR)
    cZ = array2dict(U.c[U], U.x.nxR, U.x.nxR+U.x.nxZ)
    dR = array2dict(U.c[L], 0,       L.x.nxR)
    dZ = array2dict(U.c[L], L.x.nxR, L.x.nxR+U.x.nxZ)

    PR = mat2dict(L.A[L], 0,       L.x.nxR)
    PZ = mat2dict(L.A[L], L.x.nxR, L.x.nxR+U.x.nxZ)
    QR = mat2dict(L.A[U], 0,       U.x.nxR)
    QZ = mat2dict(L.A[U], U.x.nxR, U.x.nxR+U.x.nxZ)
    s  = array2dict(L.b)
    wR = array2dict(L.c[L], 0,       L.x.nxR)
    wZ = array2dict(L.c[L], L.x.nxR, L.x.nxR+U.x.nxZ)
    '''
    mU number of upper level constraints
    mR number of upper level continuous variables
    mZ number of upper level integer variables

    nL number of lower level constraints
    nR number of lower level continuous variables
    nZ number of lower level integer variables

    AR constraint matrix for upper level problem, upper level continuous variables
    AZ constraint matrix for upper level problem, upper level integer variables
    BR constraint matrix for upper level problem, lower level continuous variables
    BZ constraint matrix for upper level problem, lower level integer variables
    r  RHS vector for upper level constraint
    cR coefficient vector for upper level objective, upper level continuous variables
    cZ coefficient vector for upper level objective, upper level integer variables
    dR coefficient vector for upper level objective, lower level continuous variables
    dZ coefficient vector for upper level objective, lower level integer variables

    PR constraint matrix for lower level problem, lower level continuous variables
    PZ constraint matrix for lower level problem, lower level integer variables
    QR constraint matrix for lower level problem, upper level continuous variables
    QZ constraint matrix for lower level problem, upper level integer variables
    s  RHS vector for lower level constraint
    wR coefficient vector for the lower level objective, lower level continuous variables
    wZ coefficient vector for the lower level objective, lower level integer variables
    '''

    #Master problem, subproblem 1, and subproblem 2 are all blocks on a Parent concrete model
    #so that parameters that are present in all three problems can be shared
    #They are mutable parameters on a parent model rather than just python dictionaries so that 
    #Pyomo constraints that would be trivially true such as 0<=0 do not cause an error

    Parent=ConcreteModel()

    Parent.mUset=RangeSet(offset, (mU-1+offset))
    Parent.nLset=RangeSet(offset, (nL-1+offset))
    Parent.nRset=RangeSet(offset, (nR-1+offset))
    Parent.nZset=RangeSet(offset, (nZ-1+offset))
    Parent.mRset=RangeSet(offset, (mR-1+offset))
    Parent.mZset=RangeSet(offset, (mZ-1+offset))

    Parent.s=Param(Parent.nLset,initialize=s,default=0,mutable=True)
    Parent.r=Param(Parent.mUset,initialize=r,default=0,mutable=True)
    Parent.cR=Param(Parent.mRset,initialize=cR,default=0,mutable=True)
    Parent.cZ=Param(Parent.mZset,initialize=cZ,default=0,mutable=True)
    Parent.dR=Param(Parent.nRset,initialize=dR,default=0,mutable=True)
    Parent.dZ=Param(Parent.nZset,initialize=dZ,default=0,mutable=True)
    Parent.wR=Param(Parent.nRset,initialize=wR,default=0,mutable=True)
    Parent.wZ=Param(Parent.nZset,initialize=wZ,default=0,mutable=True)

    Parent.AR=Param(Parent.mUset,Parent.mRset,initialize=AR, default=0,mutable=True)
    Parent.AZ=Param(Parent.mUset,Parent.mZset,initialize=AZ, default=0,mutable=True)
    Parent.BR=Param(Parent.mUset,Parent.nRset,initialize=BR, default=0,mutable=True)
    Parent.BZ=Param(Parent.mUset,Parent.nZset,initialize=BZ, default=0,mutable=True)
    Parent.PR=Param(Parent.nLset,Parent.nRset,initialize=PR, default=0,mutable=True)
    Parent.PZ=Param(Parent.nLset,Parent.nZset,initialize=PZ, default=0,mutable=True)
    Parent.QR=Param(Parent.nLset,Parent.mRset,initialize=QR, default=0,mutable=True)
    Parent.QZ=Param(Parent.nLset,Parent.mZset,initialize=QZ, default=0,mutable=True)

    Parent.zero=Param(initialize=0, mutable=True) 

    '''
    Master Problem: (P9)
        min cR*xu + cZ*yu + dR*xl0 + dZ*yl0 
        s.t. AR*xu + AZ*yu + BR*xl0 + BZ* yl0 <= r (12)
             PR*xl+PZ*yl<=s-QR*xu-QZ*yu (13)
             wR*xl0>= wR*xltilde (74)
             PR*xltilde <= s- QR*xu - QZ*yu-PZ*yl0 (75a)
             PR*pitilde>= wR (75b)
             xltilde \perp PR*pitilde - wR (76a)
             pitilde \perp s- QR*xu- QZ*yu - PR*xltilde (76b)
             PR*xlj-tj+PZ*yl<=s-QR*xu-QZ*yu for all 1<=j<=k (79)
             PR*lambdaj>=0 for all 1<=j<=k (83a)
             xlj \perp PR*lambdaj for all 1<=j<=k (83b)
             e-lambdaj >=0 for all 1<=j<=k (84a)
             tj \perp e - lambdaj for all 1<=j<=k (84b)
             lambdaj \perp s-QR*xu-QZ*yu-PR*xlj-PZ*ylj +tj for all 1<=j<=k (85)
             [e*tj=0] ==> [Constraint Block] for all 1<=j<=k (82)
             Constraint Block:
                 wR*xl0 + wZ*yl0 >= wR*xlj + wZ*ylj
                 PR*xlj <= s - QR*xu - QZ*yu - PZ*ylj
                 PR*pij>= wR
                 xlj \perp PR*pij-wR
                 pij \perp s - QR*xu - QZ*yu - PZ*ylj - PR*xlj
    '''

    Parent.Master=Block()

    Parent.Master.Y=Param(Any,within=NonNegativeIntegers,mutable=True) #Growing Parameter
    Parent.Master.xu=Var(Parent.mRset,within=NonNegativeReals,bounds=(0,M))
    Parent.Master.yu=Var(Parent.mZset,within=NonNegativeIntegers,bounds=(0,M))
    Parent.Master.xl0=Var(Parent.nRset,within=NonNegativeReals,bounds=(0,M))
    Parent.Master.yl0=Var(Parent.nZset,within=NonNegativeIntegers,bounds=(0,M))

    Parent.Master.x=Var(Any,within=NonNegativeReals,dense=False,bounds=(0,M)) #Growing variable
    Parent.Master.pi=Var(Any, within=NonNegativeReals,dense=False,bounds=(0,M)) #Growing variable

    Parent.Master.xltilde=Var(Parent.nRset,within=NonNegativeReals,bounds=(0,M))
    Parent.Master.pitilde=Var(Parent.nLset,within=NonNegativeReals,bounds=(0,M))

    Parent.Master.t=Var(Any,within=NonNegativeReals,dense=False,bounds=(0,M)) #Growing variable
    Parent.Master.lam=Var(Any,within=NonNegativeReals,dense=False,bounds=(0,M)) #Growing variable


    def Master_obj(Master):
        value=(sum(Parent.cR[j]*Parent.Master.xu[j] for j in Parent.mRset)+
               sum(Parent.cZ[j]*Parent.Master.yu[j] for j in Parent.mZset)+
               sum(Parent.dR[j]*Parent.Master.xl0[j] for j in Parent.nRset)+
               sum(Parent.dZ[j]*Parent.Master.yl0[j] for j in Parent.nZset))
        return value

    Parent.Master.Theta_star=Objective(rule=Master_obj,sense=minimize)
        
    def Master_c1(Master,i):
        value=(sum(Parent.AR[(i,j)]*Parent.Master.xu[j] for j in Parent.mRset)+
               sum(Parent.AZ[(i,j)]*Parent.Master.yu[j] for j in Parent.mZset)+
               sum(Parent.BR[(i,j)]*Parent.Master.xl0[j] for j in Parent.nRset)+
               sum(Parent.BZ[(i,j)]*Parent.Master.yl0[j] for j in Parent.nZset))
        return value - Parent.r[i] <= Parent.zero
    Parent.Master.c1=Constraint(Parent.mUset,rule=Master_c1) #(12)

    def Master_c2(Master,i):
        value=(sum(Parent.QR[(i,j)]*Parent.Master.xu[j] for j in Parent.mRset)+
               sum(Parent.QZ[(i,j)]*Parent.Master.yu[j] for j in Parent.mZset)+
               sum(Parent.PR[(i,j)]*Parent.Master.xl0[j] for j in Parent.nRset)+
               sum(Parent.PZ[(i,j)]*Parent.Master.yl0[j] for j in Parent.nZset))
        return value - Parent.s[i] <= Parent.zero
    Parent.Master.c2=Constraint(Parent.nLset,rule=Master_c2) #(13)


    def Master_c3(Master):
        l_value=sum(Parent.wR[i]*Parent.Master.xl0[i] for i in Parent.nRset)
        r_value=sum(Parent.wR[i]*Parent.Master.xltilde[i] for i in Parent.nRset)
        return l_value - r_value >= Parent.zero
    Parent.Master.c3=Constraint(rule=Master_c3) #(74)


    def Master_c4(Master,i):
        value=(sum(Parent.PR[(i,j)]*Parent.Master.xltilde[j] for j in Parent.nRset)+
               sum(Parent.PZ[(i,j)]*Parent.Master.yl0[j] for j in Parent.nZset)+
               sum(Parent.QZ[(i,j)]*Parent.Master.yu[j] for j in Parent.mZset)+
               sum(Parent.QR[(i,j)]*Parent.Master.xu[j] for j in Parent.mRset))
        return value - Parent.s[i] <= Parent.zero
    Parent.Master.c4=Constraint(Parent.nLset,rule=Master_c4) #(75a)

    def Master_c5(Master,j):
        PRpi=sum(Parent.PR[(i,j)]*Parent.Master.pitilde[i] for i in Parent.nLset)
        return PRpi - Parent.wR[j] >= Parent.zero
    Parent.Master.c5=Constraint(Parent.nRset, rule=Master_c5) #(75b)


    Parent.Master.CompBlock=Block()
    Parent.Master.CompBlock.c6=ComplementarityList(rule=(complements(Parent.Master.xltilde[j] >= 0,
                                                           sum(Parent.PR[(i,j)]*Parent.Master.pitilde[i] for i in Parent.nLset)-
                                                           Parent.wR[j] >=0) for j in Parent.nRset))
    #(76a)

    #(76b)
    Parent.Master.CompBlock.c7=ComplementarityList(rule=(complements(Parent.Master.pitilde[j] >= 0,
                                                           (Parent.s[j]-sum(Parent.QR[(j,i)]*Parent.Master.xu[i] for i in Parent.mRset) -
                                                           sum(Parent.QZ[(j,i)]*Parent.Master.yu[i] for i in Parent.mZset)-
                                                           sum(Parent.PR[(j,i)]*Parent.Master.xltilde[i] for i in Parent.nRset)-
                                                           sum(Parent.PZ[(j,i)]*Parent.Master.yl0[i] for i in Parent.nZset))>=0) for j in Parent.nLset))

    #TransformationFactory('mpec.simple_disjunction').apply_to(Parent.Master.CompBlock) #To get the complementarity not in disjunction


    Parent.Master.c_col = ConstraintList()
    Parent.Master.CompBlock2=Block(Any)

    Parent.Master.DisjunctionBlock=Block(Any)


    #Create parameters that will be updated on each iteration, initialized with coefficient vector of same size 
    Parent.xu_star=Param(Parent.mRset,initialize=cR,default=0,mutable=True)
    Parent.xl0_star=Param(Parent.nRset,initialize=wR,default=0,mutable=True)
    Parent.yu_star=Param(Parent.mZset,initialize=cZ,default=0,mutable=True)
    Parent.yl0_star=Param(Parent.nZset,initialize=wZ,default=0,mutable=True)
    theta=0
    Parent.theta=Param(initialize=theta,mutable=True)

    Parent.xl_hat=Param(Parent.nRset,initialize=0,within=NonNegativeReals,default=0,mutable=True)
    Parent.yl_hat=Param(Parent.nZset,initialize=0,within=NonNegativeIntegers,default=0,mutable=True)

    Parent.yl_star=Param(Parent.nZset, initialize=0,within=NonNegativeIntegers,default=0,mutable=True) 
    Parent.xl_star=Param(Parent.nRset, initialize=0,within=NonNegativeReals,default=0,mutable=True)
    Parent.Theta_0=Param(initialize=0,mutable=True)
    Parent.yl_arc=Param(Parent.nZset, initialize=0,within=NonNegativeIntegers,default=0,mutable=True)


    ''' Subproblem 1
        theta(xu*,yu*)=max wR*xl + wZ*yl
        s.t. PR*xl+PZ*yl <= s- QR*xu* - QZ*yu* (56)
    '''
    def sub1_obj(sub1):
        value=(sum(Parent.wR[j]*Parent.sub1.xl[j] for j in Parent.nRset)+
               sum(Parent.wZ[j]*Parent.sub1.yl[j] for j in Parent.nZset))
        return value

    def sub1_c1(sub1,i):
        value=(sum(Parent.PR[(i,j)]*Parent.sub1.xl[j] for j in Parent.nRset)+
               sum(Parent.PZ[(i,j)]*Parent.sub1.yl[j] for j in Parent.nZset)+
               sum(Parent.QR[(i,j)]*Parent.xu_star[j] for j in Parent.mRset)+
               sum(Parent.QZ[(i,j)]*Parent.yu_star[j] for j in Parent.mZset))
        return Parent.zero <= Parent.s[i] -value


    Parent.sub1=Block()   
    Parent.sub1.xl=Var(Parent.nRset, within=NonNegativeReals)
    Parent.sub1.yl=Var(Parent.nZset, within=NonNegativeIntegers)
    Parent.sub1.theta=Objective(rule=sub1_obj,sense=maximize)
    Parent.sub1.c1=Constraint(Parent.nLset,rule=sub1_c1)



    ''' Subproblem 2
        Theta_0(xu*,yu*)=min dR*xl + dZ*yl
        s.t. PR*xl+PZ*yl <= s- QR*xu* - QZ*yu* (56)
             BR*xl + BZ*yl <= r- AR*xu* - AZ*yu* (59)
             wR*xl + wZ*yl >= theta(xu*,yu*) (60)
    '''
    def sub2_obj(sub2):
        value=(sum(Parent.dR[i]*Parent.sub2.xl[i] for i in Parent.nRset)+
               sum(Parent.dZ[i]*Parent.sub2.yl[i] for i in Parent.nZset)) 
        return value

    def sub2_c1(sub2,i):
        value=(sum(Parent.PR[(i,j)]*Parent.sub2.xl[j] for j in Parent.nRset)+
               sum(Parent.PZ[(i,j)]*Parent.sub2.yl[j] for j in Parent.nZset)+
               sum(Parent.QR[(i,j)]*Parent.xu_star[j] for j in Parent.mRset)+
               sum(Parent.QZ[(i,j)]*Parent.yu_star[j] for j in Parent.mZset))
        return Parent.zero <= Parent.s[i] -value

    def sub2_c2(sub2,i):
        value=(sum(Parent.BR[(i,j)]*Parent.sub2.xl[j] for j in Parent.nRset)+
               sum(Parent.BZ[(i,j)]*Parent.sub2.yl[j] for j in Parent.nZset)+
               sum(Parent.AR[(i,j)]*Parent.xu_star[j] for j in Parent.mRset)+
               sum(Parent.AZ[(i,j)]*Parent.yu_star[j] for j in Parent.mZset))                                                                        
        return Parent.zero <= Parent.r[i] - value

    def sub2_c3(sub2):
        value=(sum(Parent.wR[j]*Parent.sub2.xl[j] for j in Parent.nRset)+
               sum(Parent.wZ[j]*Parent.sub2.yl[j] for j in Parent.nZset))
        return value - Parent.theta >= Parent.zero


    Parent.sub2=Block()
        
    Parent.sub2.xl=Var(Parent.nRset, within=NonNegativeReals)
    Parent.sub2.yl=Var(Parent.nZset, within=NonNegativeIntegers)
        
    Parent.sub2.Theta_0=Objective(rule=sub2_obj,sense=minimize)

    Parent.sub2.c1=Constraint(Parent.nLset,rule=sub2_c1) #(56)
    Parent.sub2.c2=Constraint(Parent.mUset,rule=sub2_c2) #(59)
    Parent.sub2.c3=Constraint(rule=sub2_c3)

    return Parent


def Master_add(Parent, k, epsilon): #function for adding constraints on each iteration
    #(79)
    
    Parent.Master.CompBlock2[k].c_comp=ComplementarityList()
    for i in Parent.nLset:
        r_value= (sum(Parent.PR[(i,j)]*Parent.Master.x[(j,k)] for j in Parent.nRset)-
                  Parent.Master.t[(i,k)])
        l_value= (Parent.s[i]-
                  sum(Parent.QR[(i,j)]*Parent.Master.xu[j] for j in Parent.mRset)-
                  sum(Parent.QZ[(i,j)]*Parent.Master.yu[j] for j in Parent.mZset)-
                  sum(Parent.PZ[(i,j)]*Parent.Master.Y[(j,k)] for j in Parent.nZset)) 
    
        Parent.Master.c_col.add(l_value-r_value >= Parent.zero)
    
    for i in Parent.nRset:  
        Parent.Master.c_col.add(sum(Parent.PR[(j,i)]*Parent.Master.lam[(j,k)] for j in Parent.nLset)>=Parent.zero)
        Parent.Master.CompBlock2[k].c_comp.add(complements(Parent.Master.x[(i,k)]>=0, 
                                                             sum(Parent.PR[(j,i)]*Parent.Master.lam[(j,k)] for j in Parent.nLset)>=0)) #(83)  
    
    for i in Parent.nLset:
        Parent.Master.c_col.add(1-Parent.Master.lam[(i,k)]>=Parent.zero)
        Parent.Master.CompBlock2[k].c_comp.add(complements(1-Parent.Master.lam[(i,k)]>=0,Parent.Master.t[(i,k)]>=0)) #(84)
    
    for i in Parent.nLset:
        Parent.Master.CompBlock2[k].c_comp.add(complements(Parent.Master.lam[(i,k)]>=0,(Parent.s[i]-
                                                     sum(Parent.QR[(i,j)]*Parent.Master.xu[j] for j in Parent.mRset)-
                                                     sum(Parent.QZ[(i,j)]*Parent.Master.yu[j] for j in Parent.mZset)-
                                                     sum(Parent.PZ[(i,j)]*Parent.Master.Y[(j,k)] for j in Parent.nZset)-
                                                     sum(Parent.PR[(i,j)]*Parent.Master.x[(j,k)] for j in Parent.nRset)+
                                                     Parent.Master.t[(i,k)]>=0))) #(85)
    
    
    
    #TransformationFactory('mpec.simple_disjunction').apply_to(Parent.Master.CompBlock2[k]) #To get the complementarity not in disjunction
    
    #(82) Disjunction
    Parent.Master.DisjunctionBlock[k].LH=Disjunct()
    Parent.Master.DisjunctionBlock[k].BLOCK=Disjunct()
    
    
    
    Parent.Master.DisjunctionBlock[k].LH.cons=Constraint(expr= sum(Parent.Master.t[(j,k)] for j in Parent.nLset) >= epsilon) 
    
    Parent.Master.DisjunctionBlock[k].BLOCK.cons=ConstraintList()
    
    l_value = (sum(Parent.wR[j]*Parent.Master.xl0[j] for j in Parent.nRset)+
               sum(Parent.wZ[j]*Parent.Master.yl0[j] for j in Parent.nZset))
    r_value = (sum(Parent.wR[j]*Parent.Master.x[(j,k)] for j in Parent.nRset)+
               sum(Parent.wZ[j]*Parent.Master.Y[(j,k)] for j in Parent.nZset))
    Parent.Master.DisjunctionBlock[k].BLOCK.cons.add(l_value - r_value >= 0)  #(82a) 
    
    for i in Parent.nLset:
        r_value = (Parent.s[i] - 
               sum(Parent.QR[(i,j)]*Parent.Master.xu[j] for j in Parent.mRset)-
               sum(Parent.QZ[(i,j)]*Parent.Master.yu[j] for j in Parent.mZset)-
               sum(Parent.PZ[(i,j)]*Parent.Master.Y[(j,k)] for j in Parent.nZset))
        l_value = sum(Parent.PR[(i,j)]*Parent.Master.x[(j,k)] for j in Parent.nRset)#(82b)
        
        Parent.Master.DisjunctionBlock[k].BLOCK.cons.add(r_value-l_value >= 0)
        
    for j in Parent.nRset:
        value= sum(Parent.PR[(i,j)]*Parent.Master.pi[(i,k)] for i in Parent.nLset)
        Parent.Master.DisjunctionBlock[k].BLOCK.cons.add(value >= Parent.wR[j])#(82c1)
    
    Parent.Master.DisjunctionBlock[k].BLOCK.comp1=ComplementarityList(rule=(complements(Parent.Master.pi[(j,k)]>=0, 
                                  (Parent.s[j]-
                                   sum(Parent.QR[(j,i)]*Parent.Master.xu[i] for i in Parent.mRset)-
                                   sum(Parent.QZ[(j,i)]*Parent.Master.yu[i] for i in Parent.mZset)-
                                   sum(Parent.PR[(j,i)]*Parent.Master.x[(i,k)] for i in Parent.nRset)-
                                   sum(Parent.PZ[(j,i)]*Parent.Master.Y[(i,k)] for i in Parent.nZset))>=0) for j in Parent.nLset)) #(82d)
    
    Parent.Master.DisjunctionBlock[k].BLOCK.comp2=ComplementarityList(rule=(complements(
            Parent.Master.x[(j,k)]>=0,
            sum(Parent.PR[(i,j)]*Parent.Master.pi[(i,k)] for i in Parent.nLset)-Parent.wR[j]>=0) for j in Parent.nRset)) #(82c2)
    
    
    Parent.Master.DisjunctionBlock[k].c_disj=Disjunction(expr=[Parent.Master.DisjunctionBlock[k].LH, Parent.Master.DisjunctionBlock[k].BLOCK])
    
    #TransformationFactory('mpec.simple_disjunction').apply_to(
    #    Parent.Master.DisjunctionBlock[k].BLOCK) #to get the complementarity in the disjunction
    return Parent


def UBnew(Parent):
    return sum(Parent.cR[j]*Parent.xu_star[j] for j in Parent.cR) +\
           sum(Parent.cZ[j]*Parent.yu_star[j] for j in Parent.cZ) +\
           Parent.Theta_0


def get_value(config, name, default):
    if name not in config:
        return default
    value = config[name]
    if value is None:
        return default
    return value

def check_termination(LB, UB, atol, rtol, quiet):
    if LB-UB > atol or (LB-UB)/(1+min(abs(LB),abs(UB))) > rtol: #Output
        if not quiet:
            raise RuntimeError(f'Error: Upper bound greater than lower bound after {k} iterations and {elapsed} seconds: Obj={LB} UB={UB}')

    if abs(UB-LB) <= atol or abs(UB-LB)/(1+min(abs(LB),abs(UB))) <= rtol: #Output
        return 1

    return 0


def execute_PCCG_solver(mpr, config, results):
    t = time.time()

    #These parameters can be changed for your specific problem
    epsilon = get_value(config, 'epsilon', 1e-4) #For use in disjunction approximation
    atol    = get_value(config, 'atol', 1e-8)   #absolute tolerance for UB-LB to claim convergence
    rtol    = get_value(config, 'rtol', 1e-8)   #relative tolerance for UB-LB to claim convergence
    maxit   = get_value(config, 'maxit', 5)     #Maximum number of iterations
    M       = get_value(config, 'bigm', 1e6)    #upper bound on variables
    solver  = config.mip_solver                 # MIP solver to use here
    quiet   = config.quiet                      # If True, then suppress output

    LB=-infinity
    UB=infinity
    k=0
     
    flag=0

    Parent = create_pyomo_model(mpr, M)

    bigm_xfrm = TransformationFactory('gdp.bigm')

    #Step 1: Initialization (done)
    if isinstance(solver, str):
        opt = SolverFactory(solver)
    else:
        opt = solver

    #Iteration
    while k < maxit:
        #Step 2: Solve the Master Problem
        TransformationFactory('mpec.simple_disjunction').apply_to(Parent.Master)
        bigm_xfrm.apply_to(Parent.Master) 
        res = opt.solve(Parent.Master)
        if res.solver.termination_condition !=TerminationCondition.optimal:
            raise RuntimeError("ERROR! ERROR! Master: Could not find optimal solution")

        for i in Parent.xu_star:
            Parent.xu_star[i]=Parent.Master.xu[i].value    
        for i in Parent.yu_star:
            Parent.yu_star[i]=Parent.Master.yu[i].value    
        for i in Parent.xl0_star:
            Parent.xl0_star[i]=Parent.Master.xl0[i].value
        for i in Parent.yl0_star:
            Parent.yl0_star[i]=Parent.Master.yl0[i].value

        LB=value(Parent.Master.Theta_star) 
        if not quiet:
            print(f'Iteration {k}: Master Obj={LB} UB={UB}')

        #Step 3: Terminate?
        flag = check_termination(LB, UB, atol, rtol, quiet)
        if flag:
            elapsed = time.time() - t
            break

        if not quiet:
            print("Step 4")
        #Step 4: Solve first subproblem
        results1=opt.solve(Parent.sub1) 
        
        if results1.solver.termination_condition !=TerminationCondition.optimal:
            raise RuntimeError("ERROR! ERROR! Subproblem 1: Could not find optimal solution")
        Parent.theta=value(Parent.sub1.theta)

        for i in Parent.xl_hat:
            Parent.xl_hat[i]=Parent.sub1.xl[i].value 
        for i in Parent.yl_hat:
            Parent.yl_hat[i]=int(round(Parent.sub1.yl[i].value)) 
        
        if not quiet:
            print("Step 5")
        #Step 5: Solve second subproblem
        results2=opt.solve(Parent.sub2)
        
        if results2.solver.termination_condition==TerminationCondition.optimal: #If Optimal
            for i in Parent.xl_star:
                Parent.xl_star[i]=Parent.sub2.xl[i].value
            for i in Parent.yl_star:
                Parent.yl_star[i]=int(round(Parent.sub2.yl[i].value))
                Parent.yl_arc[i]=int(round(Parent.sub2.yl[i].value))
            Parent.Theta_0=value(Parent.sub2.Theta_0)

     
            UB=min(UB,value(UBnew(Parent)))
            
        elif results2.solver.termination_condition==TerminationCondition.infeasible or results2.solver.termination_condition==TerminationCondition.infeasibleOrUnbounded: #If infeasible
            for i in Parent.yl_arc:
                Parent.yl_arc[i]=Parent.yl_hat[i]  
        else: 
             raise RuntimeError("ERROR! Unexpected termination condition for Subproblem2: %s.  Expected an infeasible or optimal solution." % str(results2.solver.termination_condition)) 
        
        if not quiet:
            print("Step 6")
        #Step 6: Add new constraints
        k = k+1
        for i in Parent.yl_arc:  #range(nZ):
            Parent.Master.Y[(i,k)]=Parent.yl_arc[i] #Make sure yl_arc is int or else Master.Y rejects
        Master_add(Parent, k, epsilon)

        if not quiet:
            print(f'Iteration {k}: Step 7 Obj={LB} UB={UB}')
        #Step 7: Loop 
        flag = check_termination(LB, UB, atol, rtol, quiet)
        if flag:
            elapsed = time.time() - t
            break

    #Output Information regarding objective and time/iterations to convergence    
    elapsed = time.time() - t
        
    results.solver.name = 'PCCG'
    results.solver_time = elapsed
    if k>= maxit:
        if not quiet:
            print('Maximum Iterations Reached')
        results.solver.termination_condition = TerminationCondition.maxIterations
    elif k< maxit and flag !=1:
        if not quiet:
            print(f'Optimal Solution Found in {k} iterations and {elapsed} seconds: Obj={UB}')
        results.solver.termination_condition = pyomo_util.pyomo2pao_termination_condition(TerminationCondition.optimal)
        results.best_feasible_objective = UB

    return Parent.Master.xu, Parent.Master.yu, Parent.Master.xl0, Parent.Master.yl0



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________
#
# pao.pyomo tests
#



#
# A solver for linear bilevel programs using
# using projected column constraint generation
# "A projection-based reformulation and decomposition algorithm for global optimization 
#  of a class of mixed integer bilevel linear programs" by Dajun Yue, Jiyao Gao, Bo Zeng, Fengqi You
#
# Adapted from an implementation by She'ifa Punla-Green at Sandia National Labs
#
#This algorithm seeks to solve the following bilevel MILP:
#    min cR*xu + cZ*yu + dR*xl0 + dZ*yl0 
#    s.t. AR*xu + AZ*yu + BR*xl0 + BZ* yl0 <= r
#     (xl0,yl0) in argmax {wR*xl+wZ*yl: PR*xl+PZ*yl<=s-QR*xu-QZ*yu}
#
import time
import numpy as np
from munch import Munch
import pyutilib
import pyomo.environ as pe
import pyomo.opt
from pyomo.common.config import ConfigBlock, ConfigValue
from pyomo.mpec import ComplementarityList, complements

import pao.common
from ..solver import Solver, LinearMultilevelSolverBase, LinearMultilevelResults
from ..repn import LinearMultilevelProblem
from ..convert_repn import convert_to_standard_form, convert_sense, convert_binaries_to_integers
from . import pyomo_util
from .pccg_solver import execute_PCCG_solver


@Solver.register(
        name='pao.mpr.PCCG',
        doc='PAO solver for Multilevel Problem Representations that define linear bilevel problems. Solver uses projected column constraint generation algorithm described by Yue et al. (2017).')
class LinearMultilevelSolver_PCCG(LinearMultilevelSolverBase):
    """
    PAO PCCG solver for linear MPRs: pao.mpr.PCCG

    This solver iteratively adds constraints to tighten a relaxation of the lower-level problem.
    """

    config = LinearMultilevelSolverBase.config()
    config.declare('mip_solver', ConfigValue(
        default='cbc',
        description="The MIP solver used by PCCG.  (default is cbc)"
        ))
    #config.declare('mip_options', ConfigValue(
    #    default=None,
    #    description="A dictionary that defines the solver options for the MIP solver.  (default is None)"))
    config.declare('bigm', ConfigValue(
        default=1e6,
        domain=float,
        description="The big-M value used to enforce complementarity conditions.       (default is 1e6)"
        ))
    config.declare('epsilon', ConfigValue(
        default=1e-4,
        domain=float,
        description="Parameter used in disjunction approximation. (default is 1e-4)"
        ))
    config.declare('atol', ConfigValue(
        default=1e-8,
        domain=float,
        description="Convergence tolerance for \|UB-LB\|. (default is 1e-8)"
        ))
    config.declare('rtol', ConfigValue(
        default=1e-8,
        domain=float,
        description="Convergence tolerance for \|UB-LB\|. (default is 1e-8)"
        ))
    config.declare('maxit', ConfigValue(
        default=None,
        domain=int,
        description="Maximum number of iterations. (default is None)"
        ))
    config.declare('quiet', ConfigValue(
        default=True,
        domain=bool,
        description="If False, then enable verbose solver output. (default is True)"
        ))

    def __init__(self, **kwds):
        super().__init__(name='pao.mpr.PCCG')

    def check_model(self, mpr):
        #
        # Confirm that the LinearMultilevelProblem is well-formed
        #
        assert (type(mpr) is LinearMultilevelProblem), "Solver '%s' can only solve a LinearMultilevelProblem" % self.name
        mpr.check()
        #
        assert (len(mpr.U.LL) == 1), "Can only solve linear bilevel problems with one lower-level"
        #
        assert (len(mpr.U.LL.LL) == 0), "Can only solve bilevel problems"

    def solve(self, mpr, options=None, **config_options):
        #
        # Error checks
        #
        self.check_model(mpr)
        #
        # Process keyword options
        #
        self._update_config(config_options)
        #
        # Start clock
        #
        start_time = time.time()

        # PCCG requires a standard form with inequalities and 
        # a maximization lower-level
        self.standard_form, soln_manager = convert_to_standard_form(mpr, inequalities=True)
        convert_sense(self.standard_form.U.LL, minimize=False)
        convert_binaries_to_integers(self.standard_form)
        
        results = LinearMultilevelResults(solution_manager=soln_manager)

        UxR, UxZ, LxR, LxZ = execute_PCCG_solver(self.standard_form, self.config, results)
        xR = {mpr.U.id:UxR, mpr.U.LL[0].id:LxR}
        xZ = {mpr.U.id:UxZ, mpr.U.LL[0].id:LxZ}

        if False:
            print("UxR")
            for i in UxR:
                print(i, UxR[i].value)
            print("UxZ")
            for i in UxZ:
                print(i, UxZ[i].value)
            print("LxR")
            for i in LxR:
                print(i, LxR[i].value)
            print("LxZ")
            for i in LxZ:
                print(i, LxZ[i].value)

        results.copy_solution(From=Munch(LxR=xR, LxZ=xZ), To=mpr)

        results.solver.wallclock_time = time.time() - start_time
        return results


pao.common.SolverAPI._generate_solve_docstring(LinearMultilevelSolver_PCCG)



import time
from pyomo.common.config import ConfigBlock, ConfigValue

from .convert import convert_pyomo2MultilevelProblem
import pao.common
import pao.mpr

Solver = pao.common.Solver


class PyomoSubmodelSolverBase(pao.common.SolverAPI):
    """
    API for solvers that optimize a Pyomo model using SubModel components.
    """

    def __init__(self, name):
        super().__init__()
        self.name = name


class PyomoSubmodelSolverBase_LBP(PyomoSubmodelSolverBase):
    """
    Define the API for solvers that optimize a Pyomo model using SubModel components
    """

    config = pao.common.solver.SolverAPI.config()
    config.declare('linearize_bigm', ConfigValue(
        default=None,
        description="The name of the big-M value used to linearize bilinear terms.  If this is not specified, then the solver will throw an error if bilinear terms exist in the model."
        ))

    def __init__(self, name, lmp_solver):
        super().__init__(name)
        self.lmp_solver = lmp_solver

    def solve(self, model, **options):
        #
        # Process keyword options
        #
        options = self._update_config(options, validate_options=False)
        solver_options = {k:self.config[k] for k in self.config}
        solver_options['load_solutions'] = True
        linearize_bigm = solver_options.pop('linearize_bigm')
        #
        # Start the clock
        #
        start_time = time.time()
        #
        # Convert the Pyomo model to a LBP
        #
        # For now, this always generates a multilevel problem with inequalities.
        # This facilitates the linearization of bilinear terms.
        #
        try:
            mp, soln_manager = convert_pyomo2MultilevelProblem(model, inequalities=True)
        except RuntimeError as err:
            print("Cannot convert Pyomo model to a multilevel problem") 
            raise
        if linearize_bigm:
            lmp, soln = pao.mpr.linearize_bilinear_terms(mp, linearize_bigm)
        else:
            lmp = mp
        #
        results = PyomoSubmodelResults(solution_manager=soln_manager)
        with pao.common.Solver(self.lmp_solver) as opt:
            lmp_results = opt.solve(lmp, **solver_options)

            self._initialize_results(results, lmp_results, model, lmp, options)
            results.solver.rc = getattr(opt, '_rc', None)
            if linearize_bigm:
                soln.copy(From=lmp, To=mp)
                results.copy(From=mp, To=model)
            else:
                results.copy(From=lmp, To=model)
            
        results.solver.wallclock_time = time.time() - start_time
        return results

    def _initialize_results(self, results, lmp_results, instance, lmp, options):
        #
        # SOLVER
        #
        solv = results.solver
        solv.name = self.name
        solv.lmp_solver = self.lmp_solver
        solv.config = self.config
        solv.termination_condition = lmp_results.solver.termination_condition
        solv.solver_time = lmp_results.solver.time
        solv.best_feasible_objective = lmp_results.solver.best_feasible_objective
        #
        # PROBLEM
        #
        prob = results.problem
        prob.name = instance.name
        prob.number_of_constraints = instance.statistics.number_of_constraints
        prob.number_of_variables = instance.statistics.number_of_variables
        prob.number_of_binary_variables = instance.statistics.number_of_binary_variables
        prob.number_of_integer_variables = instance.statistics.number_of_integer_variables
        prob.number_of_continuous_variables = instance.statistics.number_of_continuous_variables
        prob.number_of_objectives = instance.statistics.number_of_objectives
        prob.lower_bound = lmp_results.problem.lower_bound
        prob.upper_bound = lmp_results.problem.upper_bound
        prob.sense = lmp_results.problem.sense

        return results


class PyomoSubmodelResults(pao.common.Results):

    def __init__(self, solution_manager=None):
        super(pao.common.Results, self).__init__()
        self._solution_manager=solution_manager

    def copy(self, **kwds):
        self._solution_manager.copy(**kwds)

    def load_from(self, data):          # pragma: no cover
        assert (False), "load_from() is not implemented"
        self._solution_manager.load_from(data)




#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

"""
pao.pyomo.plugins.lcp
"""

import six

from pyomo.core import Block, VarList, ConstraintList, Objective,\
                       Var, Constraint, maximize, ComponentUID, Set,\
                       TransformationFactory
from pyomo.repn import generate_standard_repn
from pyomo.mpec import ComplementarityList, complements
from .transform import BaseBilevelTransformation
import logging

logger = logging.getLogger(__name__)

def create_submodel_kkt_block(instance, submodel, deterministic, fixed_upper_vars):
    """
    Add optimality conditions for the submodel

    This assumes that the original model has the form:

        min c1*x + d1*y
            A3*x <= b3
            A1*x + B1*y <= b1
            min c2*x + d2*y + x'*Q*y
                A2*x + B2*y + x'*E2*y <= b2
                y >= 0

    NOTE THE VARIABLE BOUNDS!
    """
    fixed_vars = {id(v) for v in fixed_upper_vars}
    #
    # Populate the block with the linear constraints.
    # Note that we don't simply clone the current block.
    # We need to collect a single set of equations that
    # can be easily expressed.
    #
    d2 = {}
    B2 = {}
    vtmp = {}
    utmp = {}
    sids_set = set()
    sids_list = []
    #
    block = Block(concrete=True)
    block.u = VarList() # Note: Dual variables associated to bounds in primal problem
    block.v = VarList() # Note: Dual variables associated to constraints in primal problem
    block.c1 = ConstraintList()
    block.c2 = ComplementarityList()
    block.c3 = ComplementarityList()
    #
    # Collect submodel objective terms
    #
    # TODO: detect fixed variables
    #
    for odata in submodel.component_data_objects(Objective, active=True):
        if odata.sense == maximize:
            d_sense = -1
        else:
            d_sense = 1
        #
        # Iterate through the variables in the representation
        #
        o_terms = generate_standard_repn(odata.expr, compute_values=False)
        #
        # Linear terms
        #
        for i, var in enumerate(o_terms.linear_vars):
            if id(var) in fixed_vars:
                #
                # Skip fixed upper variables
                #
                continue
            #
            # Store the coefficient for the variable.  The coefficient is
            # negated if the objective is maximized.
            #
            id_ = id(var)
            d2[id_] = d_sense * o_terms.linear_coefs[i]
            if not id_ in sids_set:
                sids_set.add(id_)
                sids_list.append(id_)
        #
        # Quadratic terms
        #
        for i, var in enumerate(o_terms.quadratic_vars):
            if id(var[0]) in fixed_vars:
                if id(var[1]) in fixed_vars:
                    #
                    # Skip fixed upper variables
                    #
                    continue
                #
                # Add the linear term
                #
                id_ = id(var[1])
                d2[id_] = d2.get(id_, 0) + d_sense * o_terms.quadratic_coefs[i] * var[0]
                if not id_ in sids_set:
                    sids_set.add(id_)
                    sids_list.append(id_)
            elif id(var[1]) in fixed_vars:
                #
                # Add the linear term
                #
                id_ = id(var[0])
                d2[id_] = d2.get(id_, 0) + d_sense * o_terms.quadratic_coefs[i] * var[1]
                if not id_ in sids_set:
                    sids_set.add(id_)
                    sids_list.append(id_)
            else:
                raise RuntimeError("Cannot apply this transformation to a problem with \
quadratic terms where both variables are in the lower level.")
        #
        # Stop after the first objective
        #
        break
    #
    # Iterate through all lower level variables, adding dual variables
    # and complementarity slackness conditions for y bound constraints
    #
    for vcomponent in instance.component_objects(Var, active=True):
        for ndx in vcomponent:
            if id(vcomponent[ndx]) in fixed_vars:
                #
                # Skip fixed upper variables
                #
                continue
            #
            # For each index, get the bounds for the variable
            #
            lb, ub = vcomponent[ndx].bounds
            if not lb is None:
                #
                # Add the complementarity slackness condition for a lower bound
                #
                v = block.v.add()
                block.c3.add(complements(vcomponent[ndx] >= lb, v >= 0))
            else:
                v = None
            if not ub is None:
                #
                # Add the complementarity slackness condition for an upper bound
                #
                w = block.v.add()
                vtmp[id(vcomponent[ndx])] = w
                block.c3.add(complements(vcomponent[ndx] <= ub, w >= 0))
            else:
                w = None
            if not (v is None and w is None):
                #
                # Record the variables for which complementarity slackness conditions
                # were created.
                #
                id_ = id(vcomponent[ndx])
                vtmp[id_] = (v, w)
                if not id_ in sids_set:
                    sids_set.add(id_)
                    sids_list.append(id_)
    #
    # Iterate through all constraints, adding dual variables and
    # complementary slackness conditions (for inequality constraints)
    #
    for cdata in submodel.component_data_objects(Constraint, active=True):
        if cdata.equality:
            # Don't add a complementary slackness condition for an equality constraint
            u = block.u.add()
            utmp[id(cdata)] = (None, u)
        else:
            if not cdata.lower is None:
                #
                # Add the complementarity slackness condition for a greater-than inequality
                #
                u = block.u.add()
                block.c2.add(complements(- cdata.body <= - cdata.lower, u >= 0))
            else:
                u = None
            if not cdata.upper is None:
                #
                # Add the complementarity slackness condition for a less-than inequality
                #
                w = block.u.add()
                block.c2.add(complements(cdata.body <= cdata.upper, w >= 0))
            else:
                w = None
            if not (u is None and w is None):
                utmp[id(cdata)] = (u, w)
        #
        # Store the coefficients for the constraint variables that are not fixed
        #
        c_terms = generate_standard_repn(cdata.body, compute_values=False)
        #
        # Linear terms
        #
        for i, var in enumerate(c_terms.linear_vars):
            if id(var) in fixed_vars:
                continue
            id_ = id(var)
            B2.setdefault(id_, {}).setdefault(id(cdata), c_terms.linear_coefs[i])
            if not id_ in sids_set:
                sids_set.add(id_)
                sids_list.append(id_)
        #
        # Quadratic terms
        #
        for i, var in enumerate(c_terms.quadratic_vars):
            if id(var[0]) in fixed_vars:
                if id(var[1]) in fixed_vars:
                    continue
                id_ = id(var[1])
                if id_ in B2:
                    B2[id_][id(cdata)] = c_terms.quadratic_coefs[i] * var[0]
                else:
                    B2.setdefault(id_, {}).setdefault(id(cdata),
                                                      c_terms.quadratic_coefs[i] * var[0])
                if not id_ in sids_set:
                    sids_set.add(id_)
                    sids_list.append(id_)
            elif id(var[1]) in fixed_vars:
                id_ = id(var[0])
                if id_ in B2:
                    B2[id_][id(cdata)] = c_terms.quadratic_coefs[i] * var[1]
                else:
                    B2.setdefault(id_, {}).setdefault(id(cdata),
                                                      c_terms.quadratic_coefs[i] * var[1])
                if not id_ in sids_set:
                    sids_set.add(id_)
                    sids_list.append(id_)
            else:
                raise RuntimeError("Cannot apply this transformation to a problem with \
quadratic terms where both variables are in the lower level.")
    #
    # Generate stationarity equations
    #
    tmp__ = (None, None)
    for vid in sids_list:
        exp = d2.get(vid, 0)
        #
        lb_dual, ub_dual = vtmp.get(vid, tmp__)
        if vid in vtmp:
            if not lb_dual is None:
                exp -= lb_dual             # dual for variable lower bound
            if not ub_dual is None:
                exp += ub_dual             # dual for variable upper bound
        #
        B2_ = B2.get(vid, {})
        utmp_keys = list(utmp.keys())
        if deterministic:
            utmp_keys.sort(key=lambda x: utmp[x][0].local_name\
                           if utmp[x][1] is None else utmp[x][1].local_name)
        for uid in utmp_keys:
            if uid in B2_:
                lb_dual, ub_dual = utmp[uid]
                if not lb_dual is None:
                    exp -= B2_[uid] * lb_dual
                if not ub_dual is None:
                    exp += B2_[uid] * ub_dual
        if type(exp) in six.integer_types or type(exp) is float:
            # TODO: Annotate the model as unbounded
            raise IOError("Unbounded variable without side constraints")
        block.c1.add(exp == 0)
    return block


@TransformationFactory.register('pao.pyomo.linear_mpec',
                                doc="Generate a linear MPEC from the optimality conditions \
of the submodel")
class LinearComplementarityBilevelTransformation(BaseBilevelTransformation):
    """
    This transformation creates a block using a SubModel object,
    which contains constraints describing the optimality conditions for that
    submodel.
    """

    def _apply_to(self, model, **kwds):
        deterministic = kwds.pop('deterministic', False)
        submodel_name = kwds.pop('submodel', None)

        #
        # Process options
        #
        self._preprocess('pao.pyomo.linear_mpec', model)

        def _sub_transformation(model, sub, key):
            model.reclassify_component_type(sub, Block)
            #
            # Create a block with optimality conditions
            #
            setattr(model, key +'_kkt',
                    create_submodel_kkt_block(model, sub, deterministic,
                                              self.fixed_vardata[key]))
            model._transformation_data['pao.pyomo.linear_mpec'].submodel_cuid =\
                ComponentUID(sub)
            model._transformation_data['pao.pyomo.linear_mpec'].block_cuid =\
                ComponentUID(getattr(model, key +'_kkt'))
            #
            # Disable the original submodel and
            #
            for data in sub.component_map(active=True).values():
                if not isinstance(data, Var) and not isinstance(data, Set):
                    data.deactivate()

        if not submodel_name is None:
            lookup = {value: key for key, value in self.submodel}
            sub = getattr(model,submodel_name)
            if sub:
                _sub_transformation(model, sub, lookup[sub])
            return

        for key, sub in self.submodel.items():
            _sub_transformation(model, sub, key)





import sys
import subprocess
import io

from pyutilib.misc import Bunch
from pyutilib.subprocess import run_command



def run_shellcmd(cmd, *, env=None, tee=False, time_limit=None):
    ostr = io.StringIO()
    rc, log = run_command(cmd, tee=tee, env=env, timelimit=time_limit,
            stdout=ostr,
            stderr=ostr)

    return Bunch(rc=rc, log=ostr.getvalue())




#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

import pyomo.opt

safe_termination_conditions = [
    pyomo.opt.TerminationCondition.maxTimeLimit,
    pyomo.opt.TerminationCondition.maxIterations,
    pyomo.opt.TerminationCondition.minFunctionValue,
    pyomo.opt.TerminationCondition.minStepLength,
    pyomo.opt.TerminationCondition.globallyOptimal,
    pyomo.opt.TerminationCondition.locallyOptimal,
    pyomo.opt.TerminationCondition.feasible,
    pyomo.opt.TerminationCondition.optimal,
    pyomo.opt.TerminationCondition.maxEvaluations,
    pyomo.opt.TerminationCondition.other
]

optimal_termination_conditions = [
    pyomo.opt.TerminationCondition.globallyOptimal,
    pyomo.opt.TerminationCondition.locallyOptimal,
    pyomo.opt.TerminationCondition.optimal
]

def _check_termination_condition(results):
    # do we want to be more restrictive of termination conditions?
    # do we want to have different behavior for sub-optimal termination?
    if results.solver.termination_condition not in optimal_termination_conditions:
        return False
    else:
        return True
